import asyncio
import dataclasses as dc
import logging
import traceback
import time
from typing import Optional, Callable

import click

import statey as st
from statey.cli.inspect import color_for_status
from statey.task import Task, TaskStatus


class TaskWrapper(Task):
    """
    Base class for tasks. A task is a unit of computation
    """

    task: Task

    @property
    def description(self) -> Optional[str]:
        return self.task.description

    def always_eager(self) -> bool:
        return self.task.always_eager()

    def is_metatask(self) -> bool:
        return self.task.is_metatask()


@dc.dataclass(frozen=True)
class TaskLogger(TaskWrapper):
    """
    Log events and heartbeats for tasks
    """

    task: Task
    key: str
    heartbeat_interval: float = 15.0

    async def run(self) -> None:
        start = time.time()
        asyncio_task = asyncio.ensure_future(self.task.run())

        while True:
            shielded = asyncio.shield(asyncio_task)
            try:
                await asyncio.wait_for(shielded, timeout=self.heartbeat_interval)
            except asyncio.TimeoutError:
                click.echo(
                    f"Task {click.style(self.key, fg='yellow')} still running "
                    f"({time.time() - start:.2f}s elapsed)."
                )
                continue
            except asyncio.CancelledError:
                asyncio_task.cancel()
                raise
            return


class ExecutorLoggingPlugin:
    """
    Plugin to log signals when caught
    """

    def __init__(
        self, show_metatasks: bool = False, heartbeat_interval: float = 15.0
    ) -> None:
        self.task_starts = {}
        self.show_metatasks = show_metatasks
        self.heartbeat_interval = heartbeat_interval

    @st.hookimpl
    def caught_signal(
        self, signals: int, max_signals: int, executor: "TaskGraphExecutor"
    ) -> None:
        if signals >= max_signals:
            click.secho(
                f"Caught {signals} signals(s) >= max_signals = {max_signals}. "
                f"Exiting forcefully, this may cause data loss.",
                fg="red",
            )
        else:
            color = "yellow" if signals == 1 else "red"
            click.secho(
                f"Caught {signals} signal(s), exiting gracefully. Additional signals may cause data loss.",
                fg=color,
            )

    def should_show(self, task: Task) -> bool:
        return self.show_metatasks or not task.is_metatask()

    @st.hookimpl
    def before_run(
        self, key: str, task: st.Task, executor: "TaskGraphExecutor"
    ) -> None:
        self.task_starts[key] = time.time()
        if self.should_show(task):
            click.echo(f'Task {click.style(key, fg="yellow")} is running.')

    @st.hookimpl
    def after_run(
        self, key: str, task: st.Task, status: TaskStatus, executor: "TaskGraphExecutor"
    ) -> None:
        duration = time.time() - self.task_starts[key]
        color = color_for_status(status)
        styled_status = click.style(status.name, fg=color, bold=True)
        if self.should_show(task):
            click.echo(
                f'Task {click.style(key, fg="yellow")} finished after {duration:.2f}s with status {styled_status}.'
            )

    @st.hookimpl
    def task_wrapper(
        self, key: str, task: st.Task, executor: "TaskGraphExecutor"
    ) -> Callable[[st.Task], st.Task]:
        if not self.should_show(task):
            return None
        return lambda input_task: TaskLogger(input_task, key, self.heartbeat_interval)


class CLILoggingHandler(logging.Handler):
    """"""

    def __init__(self, fulltrace: bool = False) -> None:
        """"""
        self.fulltrace = fulltrace
        super().__init__()

    def emit(self, record):

        printer = click.echo

        if record.levelno == logging.WARNING:
            printer = lambda x: click.secho(x, fg="yellow")
        elif record.levelno >= logging.ERROR:
            printer = lambda x: click.secho(x, fg="red")

        printer(record.getMessage())

        if record.exc_info is not None and self.fulltrace:
            formatted = traceback.format_exception(*record.exc_info)
            click.secho("".join(formatted), fg="red")



class StorageError(Exception):
    """
    Base class for errors relating to storing or loading resource
    graphs
    """


class StateLoadError(StorageError):
    """
    Specifically errors relate to loading a graph from storage
    """


class StateDumpError(StorageError):
    """
    Specifically errors related to dumping a graph to storage
    """



import dataclasses as dc
from typing import Optional, Any, Dict, Sequence

import statey as st
from statey.ext.pulumi import PulumiProvider
from statey.syms import utils


class APIMixin:
    """
    Common behavior for API classes
    """

    def path(self) -> str:
        return ".".join(filter(None, [self.parent and self.parent.path(), self.name]))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path()})"


class ProvidersAPI(APIMixin):
    """
    A simple interface for creating resource states fluently
    """

    def __init__(self, registry: Optional[st.Registry] = None) -> None:
        if registry is None:
            registry = st.registry
        self.registry = registry
        self.name = None
        self.parent = None
        self._providers = set()

    def get_provider(self, name: str) -> st.Provider:
        """
        Get the pulumi provider with the given name
        """
        api = ProviderAPI(name, self.registry.get_provider("pulumi/" + name), self)
        self._providers.add(name)
        return api

    def __dir__(self) -> Sequence[str]:
        """"""
        return dir(super()) + list(self._providers)

    def __getattr__(self, attr: str) -> "API_Provider":
        """"""
        try:
            return getattr(super(), attr)
        except AttributeError:
            if attr.startswith("__"):
                raise

        return self.get_provider(attr)


providers = ProvidersAPI()


class ProviderAPI(APIMixin):
    """"""

    def __init__(self, name: str, provider: PulumiProvider, parent: Optional[Any] = None) -> None:
        self.name = name
        self.provider = provider
        self.parent = parent
        self.modules = None

    def _build_modules(self, provider: PulumiProvider) -> Dict[str, Dict[str, str]]:
        names = {}

        for resource in self.provider.schema.resources:
            _, mod_name, typ = resource.split(":")
            # Might end up being a bug, but using the full names is ugly :(
            root_mod, *_ = mod_name.split("/", 1)
            names.setdefault(root_mod, {})[typ] = resource

        return names

    def _get_modules(self) -> Dict[str, Dict[str, str]]:
        if self.modules is None:
            self.modules = self._build_modules(self.provider)
        return self.modules

    def get_module(self, name: str) -> "ModuleAPI":
        """"""
        if name not in self._get_modules():
            raise KeyError(name)
        return ModuleAPI(name, self.modules[name], self.provider, self)

    def __dir__(self) -> Sequence[str]:
        """"""
        return dir(super()) + list(self._get_modules())

    def __getattr__(self, attr: str) -> "ModuleAPI":
        """"""
        try:
            return getattr(super(), attr)
        except AttributeError:
            if attr.startswith("__"):
                raise

        return self.get_module(attr)


class ModuleAPI(APIMixin):
    """"""

    def __init__(
        self,
        name: str,
        resources: Dict[str, Dict[str, str]],
        provider: PulumiProvider,
        parent: Any,
    ) -> None:
        self.name = name
        self.provider = provider
        self.parent = parent
        self.resources = resources

    def get_resource(self, name: str) -> st.BoundState:
        """"""
        if name not in self.resources:
            raise KeyError(name)
        resource_name = self.resources[name]
        resource = self.provider.get_resource(resource_name)
        return resource

    def __dir__(self) -> Sequence[str]:
        """"""
        return dir(super()) + list(self.resources)

    def __getattr__(self, attr: str) -> "ProviderAPI":
        """"""
        try:
            return getattr(super(), attr)
        except AttributeError:
            if attr.startswith("__"):
                raise

        return self.get_resource(attr)



import abc
import dataclasses as dc
import operator
from typing import (
    Sequence,
    Any,
    Iterator,
    Dict,
    Type as PyType,
    Tuple,
    Callable,
    Union,
    Optional,
)

import statey as st
from statey.syms import types, utils, impl, Object
from statey.syms.path import PathParser


PathLike = Union[Sequence[Any], str]


def get_path_from_pathlike(
    path_like: PathLike, path_parser: PathParser
) -> Sequence[Any]:
    if isinstance(path_like, str):
        path_like = path_parser.split(path_like)
    return path_like


@dc.dataclass
class DiffConfig(utils.Cloneable):
    """
    Simple class to customize behavior when diffing
    """

    path_comparisons: Dict[Sequence[Any], Callable[[Any, Any], bool]] = dc.field(
        default_factory=dict
    )
    path_parser: PathParser = dc.field(default_factory=PathParser)

    def set_comparison(self, path: PathLike, func: Callable[[Any, Any], bool]) -> None:
        """
        Modify this config in place, setting the comparison for the given path
        """
        self.path_comparisons[get_path_from_pathlike(path, self.path_parser)] = func

    def get_explicit_comparison(
        self, path: PathLike
    ) -> Optional[Callable[[Any, Any], bool]]:
        """
        Get the comparison function for a given path, _only_ if it has been explicitly
        set using set_comparison(). Otherwise, return None
        """
        path = get_path_from_pathlike(path, self.path_parser)
        if path in self.path_comparisons:
            return self.path_comparisons[path]
        return None

    def get_comparison(self, path: PathLike) -> Callable[[Any, Any], bool]:
        """
        Get the comparison function for a given path
        """
        comp = self.get_explicit_comparison(path)
        if comp is None:
            return operator.eq

        def wrapped(old, new):
            res = comp(old, new)
            if res is NotImplemented:
                return old == new
            return res

        return wrapped


@dc.dataclass(frozen=True)
class DiffComponent:
    """
    A diff represents the differences between two values. Either left or right can contain unknowns,
    however neither should contain objects (i.e. they should be either fully resolve or the result of
    Session.resolve() with allow_unknowns=True)
    """

    left: Any
    right: Any
    differ: "Differ"
    path: Sequence[Any] = ()

    def __bool__(self) -> bool:
        try:
            next(self.flatten())
        except StopIteration:
            return True
        return False

    def left_is_unknown(self) -> bool:
        """
        Indicate whether the left side of this diff is an `Unknown` instance
        """
        return isinstance(self.left, Object) and isinstance(
            self.left._impl, impl.Unknown
        )

    def right_is_unknown(self) -> bool:
        """
        Indicate whether the right side of this diff is an `Unknown` instance
        """
        return isinstance(self.right, Object) and isinstance(
            self.right._impl, impl.Unknown
        )

    def flatten(self, config: Optional[DiffConfig] = None) -> Iterator["DiffComponent"]:
        """
        Flatten this diff into any non-empty sub-diffs
        """
        return self.differ.flatten(self, config)


@dc.dataclass(frozen=True)
class Diff:
    """
    A flattened out diff object, suitable for actual inspection
    """

    components: Sequence[DiffComponent]
    path_parser: PathParser

    def _get_path(self, path_like: PathLike) -> Sequence[Any]:
        if isinstance(path_like, str):
            path_like = self.path_parser.split(path_like)
        return path_like

    def __iter__(self) -> Iterator[DiffComponent]:
        """
        Iterate through component diffs
        """
        return iter(self.components)

    def __bool__(self) -> bool:
        """
        Indicate whether this flat diff is empty
        """
        return bool(self.components)

    def __getitem__(self, path_like: PathLike) -> DiffComponent:
        """
        Get a diff by path
        """
        comps = get_path_from_pathlike(path_like, self.path_parser)
        for diff in self.components:
            if comps == tuple(diff.path):
                return diff
        raise KeyError(comps)

    def __contains__(self, path_like: PathLike) -> bool:
        """
        Indicate whether a diff at the given path exists
        """
        comps = get_path_from_pathlike(path_like, self.path_parser)
        for diff in self.components:
            if comps == tuple(diff.path)[: len(comps)]:
                return True
        return False


class Differ(abc.ABC):
    """
    A differ provides a simple way to get a flat set of differences from a potentially
    complex nested object
    """

    type: types.Type

    def config(self) -> DiffConfig:
        """
        Helper function to grab a fresh
        """
        return DiffConfig()

    @abc.abstractmethod
    def flatten(
        self, diff: DiffComponent, config: Optional[DiffConfig] = None
    ) -> Iterator[DiffComponent]:
        """
        Flatten this diff into any non-empty sub-diffs
        """
        raise NotImplementedError

    def diff(
        self,
        left: Any,
        right: Any,
        session: Optional["Session"] = None,
        config: Optional[DiffConfig] = None,
    ) -> Diff:
        """
        Construct a Diff object from this differ and the given data
        """
        if session is not None:
            left_obj = Object(left, self.type, session.ns.registry)
            left = session.resolve(left_obj, decode=False, allow_unknowns=True)
            right_obj = Object(right, self.type, session.ns.registry)
            right = session.resolve(right_obj, decode=False, allow_unknowns=True)

        if config is None:
            config = self.config()

        diff = DiffComponent(left, right, self)
        diffs = list(diff.flatten(config))

        return Diff(diffs, config.path_parser)


@dc.dataclass(frozen=True)
class ValueDiffer(Differ):
    """
    Differ for a single value
    """

    type: types.Type

    def is_empty(self, diff: "Diff", config: DiffConfig) -> bool:
        if diff.left_is_unknown() or diff.right_is_unknown():
            return False
        comp = config.get_comparison(diff.path)
        return comp(diff.left, diff.right)

    def flatten(
        self, diff: DiffComponent, config: Optional[DiffConfig] = None
    ) -> Iterator[DiffComponent]:
        if config is None:
            config = self.config()

        if not self.is_empty(diff, config):
            yield diff

    @classmethod
    @st.hookimpl
    def get_differ(cls, type: types.Type, registry: st.Registry) -> "Differ":
        if not isinstance(
            type,
            (types.IntegerType, types.FloatType, types.BooleanType, types.StringType),
        ):
            return None
        return cls(type)


@dc.dataclass(frozen=True)
class ArrayDiffer(Differ):
    """
    Differ for arrays
    """

    type: types.Type
    element_differ: Differ

    def element_diffs(self, diff: DiffComponent) -> Sequence[DiffComponent]:
        """
        Construct diffs for each of the elements of the givn diff
        """
        out = []

        for idx, (left_item, right_item) in enumerate(zip(diff.left, diff.right)):
            sub_diff = DiffComponent(
                differ=self.element_differ,
                left=left_item,
                right=right_item,
                path=tuple(diff.path) + (idx,),
            )
            out.append(sub_diff)

        if len(diff.left) > len(diff.right):
            for idx, left_item in enumerate(diff.left[len(diff.right) :]):
                sub_diff = DiffComponent(
                    differ=self.element_differ,
                    left=left_item,
                    right=None,
                    path=tuple(diff.path) + (idx + len(diff.right),),
                )
                out.append(sub_diff)

        if len(diff.right) > len(diff.left):
            for idx, right_item in enumerate(diff.right[len(diff.left) :]):
                sub_diff = DiffComponent(
                    differ=self.element_differ,
                    left=None,
                    right=right_item,
                    path=tuple(diff.path) + (idx + len(diff.left),),
                )
                out.append(sub_diff)

        return out

    def flatten(
        self, diff: DiffComponent, config: Optional[DiffConfig] = None
    ) -> Iterator[DiffComponent]:
        """
        Yield the diffs for each element in left and right. This assumes these arrays are of the
        same length
        """
        if config is None:
            config = self.config()

        if diff.left_is_unknown() or diff.right_is_unknown():
            yield diff
            return

        explicit_comp = config.get_explicit_comparison(diff.path)
        if explicit_comp is not None:
            res = explicit_comp(diff.left, diff.right)
            if not res:
                yield diff
            if res is not NotImplemented:
                return

        if diff.left is None or diff.right is None:
            if diff.left is None and diff.right is None:
                return
            yield diff
            return

        if len(diff.left) != len(diff.right):
            yield diff
            return

        for diff in self.element_diffs(diff):
            yield from diff.flatten(config)

    @classmethod
    @st.hookimpl
    def get_differ(cls, type: types.Type, registry: st.Registry) -> "Differ":
        if not isinstance(type, types.ArrayType):
            return None
        element_differ = registry.get_differ(type.element_type)
        return cls(type, element_differ)


@dc.dataclass(frozen=True)
class StructDiffer(Differ):
    """
    Differ for arrays
    """

    type: types.Type
    field_differs: Dict[str, Differ]

    def field_diffs(self, diff: DiffComponent) -> Dict[str, DiffComponent]:
        """
        Construct diffs for each field given the input diff
        """
        out = {}

        for key, differ in self.field_differs.items():
            left_value = diff.left[key]
            right_value = diff.right[key]
            out[key] = DiffComponent(
                differ=differ,
                left=left_value,
                right=right_value,
                path=tuple(diff.path) + (key,),
            )

        return out

    def flatten(
        self, diff: DiffComponent, config: Optional[DiffConfig] = None
    ) -> Iterator[DiffComponent]:
        if config is None:
            config = self.config()

        if diff.left_is_unknown() or diff.right_is_unknown():
            yield diff
            return

        explicit_comp = config.get_explicit_comparison(diff.path)
        if explicit_comp is not None:
            res = explicit_comp(diff.left, diff.right)
            if not res:
                yield diff
            if res is not NotImplemented:
                return

        if diff.left is None or diff.right is None:
            if diff.left is None and diff.right is None:
                return
            yield diff
            return

        for diff in self.field_diffs(diff).values():
            yield from diff.flatten(config)

    @classmethod
    @st.hookimpl
    def get_differ(cls, type: types.Type, registry: st.Registry) -> "Differ":
        if not isinstance(type, types.StructType):
            return None
        field_differs = {}
        for field in type.fields:
            field_differs[field.name] = registry.get_differ(field.type)
        return cls(type, field_differs)


@dc.dataclass(frozen=True)
class MapDiffer(Differ):
    """
    Differ for arrays
    """

    type: types.Type
    value_differ: Differ

    def element_diffs(self, diff: DiffComponent) -> Sequence[DiffComponent]:
        """
        Construct diffs for each of the elements of the givn diff
        """
        out = []

        left_keys = set(diff.left)
        right_keys = set(diff.right)

        for key in left_keys - right_keys:
            sub_diff = DiffComponent(
                differ=self.value_differ,
                left=diff.left[key],
                right=None,
                path=tuple(diff.path) + (key,),
            )
            out.append(sub_diff)

        for key in right_keys - left_keys:
            sub_diff = DiffComponent(
                differ=self.value_differ,
                left=diff.right[key],
                right=None,
                path=tuple(diff.path) + (key,),
            )
            out.append(sub_diff)

        for key in left_keys & right_keys:
            sub_diff = DiffComponent(
                differ=self.value_differ,
                left=diff.left[key],
                right=diff.right[key],
                path=tuple(diff.path) + (key,),
            )
            out.append(sub_diff)

        return out

    def flatten(
        self, diff: DiffComponent, config: Optional[DiffConfig] = None
    ) -> Iterator[DiffComponent]:
        """
        Yield the diffs for each element in left and right. This assumes these arrays are of the
        same length
        """
        if config is None:
            config = self.config()

        if diff.left_is_unknown() or diff.right_is_unknown():
            yield diff
            return

        explicit_comp = config.get_explicit_comparison(diff.path)
        if explicit_comp is not None:
            res = explicit_comp(diff.left, diff.right)
            if not res:
                yield diff
            if res is not NotImplemented:
                return

        if diff.left is None or diff.right is None:
            if diff.left is None and diff.right is None:
                return
            yield diff
            return

        if len(diff.left) != len(diff.right):
            yield diff
            return

        for diff in self.element_diffs(diff):
            yield from diff.flatten(config)

    @classmethod
    @st.hookimpl
    def get_differ(cls, type: types.Type, registry: st.Registry) -> "Differ":
        if not isinstance(type, types.MapType):
            return None
        value_differ = registry.get_differ(type.value_type)
        return cls(type, value_differ)


DIFFER_HOOKS = [ValueDiffer, ArrayDiffer, StructDiffer, MapDiffer]


def register(registry: Optional["Registry"] = None) -> None:
    if registry is None:
        registry = st.registry

    for plugin in DIFFER_HOOKS:
        registry.register(plugin)



import abc
import dataclasses as dc
from typing import Dict, Any, Callable, Optional

import statey as st
from statey.syms import types, utils, func
from statey.syms.object_ import Object


class Caster(abc.ABC):
    """
    A caster defines behavior to cast a value from one type to another
    """

    type: types.Type

    @abc.abstractmethod
    def cast_value(self, value: Any) -> Any:
        """
        Cast a concrete value between types rather than an object
        """
        raise NotImplementedError

    @abc.abstractmethod
    def cast(self, obj: Object) -> Object:
        """
        Cast the given object to this caster's type
        """
        raise NotImplementedError


class ValueCaster(Caster):
    """
    Cast from one type to another using a native python function
    """

    type: types.Type

    def cast(self, obj: Object) -> Object:
        func_type = utils.single_arg_function_type(obj._type, self.type)
        function = func.NativeFunction(
            type=func_type,
            func=self.cast_value,
            name=f"cast[{obj._type} => {self.type}]",
        )
        return obj._inst.map(function)


@dc.dataclass(frozen=True)
class FunctionalCaster(ValueCaster):
    """
    ValueCaster using a function as the cast_value() method
    """

    type: types.Type
    cast_value: Callable[[Any], Any] = dc.field()


@dc.dataclass(frozen=True)
class ForceCaster(Caster):
    """
    Caster that simply replaces the type of a given object. This will cause the object
    to be encoded as that type
    """

    type: types.Type

    def cast_value(self, value: Any) -> Any:
        return value

    def cast(self, obj: Object) -> Object:
        return Object(obj._impl, self.type, obj._registry)


@dc.dataclass(frozen=True)
class ArrayElementCaster(ValueCaster):
    """
    Caster for array types
    """

    type: types.ArrayType
    element_caster: Caster

    def cast_value(self, value: Any) -> Any:
        if value is None:
            return None
        return [self.element_caster.cast_value(item) for item in value]


@dc.dataclass(frozen=True)
class MapElementCaster(ValueCaster):
    """
    Caster for map types
    """

    type: types.MapType
    key_caster: Caster
    value_caster: Caster

    def cast_value(self, value: Any) -> Any:
        if value is None:
            return None
        return {
            self.key_caster.cast_value(key): self.value_caster.cast_value(val)
            for key, val in value.items()
        }


@dc.dataclass(frozen=True)
class StructCaster(Caster):
    """
    Caster for struct types
    """

    type: types.StructType
    field_casters: Dict[str, Caster]

    def cast_value(self, value: Any) -> Any:
        if value is None:
            return None
        return {
            field: caster.cast_value(value[field]) for field, caster in value.items()
        }

    def cast(self, obj: Object) -> Object:
        out = {}
        for field in self.type.fields:
            caster = self.field_casters[field.name]
            out[field.name] = caster.cast(obj[field.name])
        return Object(out, self.type, obj._registry)


@dc.dataclass(frozen=True)
class PredicateCastingPlugin:
    """
    Provide a (from_type, to_type) => bool predicate and a caster factory
    """

    factory: Callable[[types.Type, types.Type, "Registry"], Optional[Caster]]

    @st.hookimpl
    def get_caster(
        self, from_type: types.Type, to_type: types.Type, registry: "Registry"
    ) -> Caster:
        return self.factory(from_type, to_type, registry)


def quasi_equal_cast(
    from_type: types.Type, to_type: types.Type, registry: "Registry"
) -> Optional[Caster]:
    """
    Predicate returning True when the types are the same other than metadata
    """
    if (
        from_type.with_meta(to_type.meta) == to_type
        or from_type.with_nullable(True).with_meta(to_type.meta) == to_type
    ):
        return ForceCaster(to_type)
    return None


def array_cast(
    from_type: types.Type, to_type: types.Type, registry: "Registry"
) -> Optional[Caster]:
    """
    Simple identity caster for arrays
    """
    if not isinstance(from_type, types.ArrayType) or not isinstance(
        to_type, types.ArrayType
    ):
        return None

    element_caster = None
    if from_type.element_type != to_type.element_type:
        try:
            element_caster = registry.get_caster(
                from_type.element_type, to_type.element_type
            )
        except st.exc.NoCasterFound:
            return None

    with_from_element = to_type.with_element_type(from_type.element_type)
    if from_type == with_from_element.with_meta(
        from_type.meta
    ) or from_type == with_from_element.with_nullable(False).with_meta(from_type.meta):
        if element_caster is None or isinstance(element_caster, ForceCaster):
            return ForceCaster(to_type)

        return ArrayElementCaster(to_type, element_caster)

    return None


def struct_cast(
    from_type: types.Type, to_type: types.Type, registry: "Registry"
) -> Optional[Caster]:
    """
    Simple identity caster for structs
    """
    if not isinstance(from_type, types.StructType) or not isinstance(
        to_type, types.StructType
    ):
        return None
    if {field.name for field in from_type.fields} != {
        field.name for field in to_type.fields
    }:
        return None

    field_casters = None
    if {field.name: field for field in from_type.fields} != {
        field.name: field for field in to_type.fields
    }:
        field_casters = {}
        for field in to_type.fields:
            try:
                field_caster = registry.get_caster(
                    from_type[field.name].type, field.type
                )
            except st.exc.NoCasterFound:
                return None
            field_casters[field.name] = field_caster

    with_from_fields = to_type.with_fields(from_type.fields)
    if from_type == with_from_fields.with_meta(
        from_type.meta
    ) or from_type == with_from_fields.with_nullable(False).with_meta(from_type.meta):
        if field_casters is None or all(
            isinstance(caster, ForceCaster) for caster in field_casters.values()
        ):
            return ForceCaster(to_type)

        return StructCaster(to_type, field_casters)

    return None


def map_cast(
    from_type: types.Type, to_type: types.Type, registry: "Registry"
) -> Optional[Caster]:
    """
    Simple identity caster for maps
    """
    if not isinstance(from_type, types.MapType) or not isinstance(
        to_type, types.MapType
    ):
        return None

    key_caster = None
    if from_type.key_type != to_type.key_type:
        try:
            key_caster = registry.get_caster(from_type.key_type, to_type.value_type)
        except st.exc.NoCasterFound:
            return None

    value_caster = None
    if from_type.value_type != to_type.value_type:
        try:
            value_caster = registry.get_caster(from_type.value_type, to_type.value_type)
        except st.exc.NoCasterFound:
            return None

    with_same_subtypes = to_type.with_key_type(from_type.key_type).with_value_type(
        from_type.value_type
    )

    if from_type == with_same_subtypes.with_meta(
        from_type.meta
    ) or from_type == with_same_subtypes.with_nullable(False).with_meta(from_type.meta):
        if (key_caster is None and value_caster is None) or (
            isinstance(key_caster, ForceCaster)
            and isinstance(value_caster, ForceCaster)
        ):
            return ForceCaster(to_type)

        if key_caster is None:
            key_caster = ForceCaster(to_type.key_type)

        if value_caster is None:
            value_caster = ForceCaster(to_type.value_type)

        return MapElementCaster(to_type, key_caster, value_caster)

    return None


def number_cast(
    from_type: types.Type, to_type: types.Type, registry: "Registry"
) -> Optional[Caster]:
    """
    Caster allowing different numerical types to be cast to one another.
    """
    if (
        type(from_type) != type(to_type)
        and isinstance(from_type, types.NumberType)
        and isinstance(to_type, types.NumberType)
    ):
        to_as_from = to_type.with_nullable(from_type.nullable).with_meta(from_type.meta)
        return registry.get_caster(to_as_from, to_type)
    return None


PLUGINS = [
    PredicateCastingPlugin(number_cast),
    PredicateCastingPlugin(array_cast),
    PredicateCastingPlugin(struct_cast),
    PredicateCastingPlugin(map_cast),
    PredicateCastingPlugin(quasi_equal_cast),
]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register casters in the default registry
    """
    if registry is None:
        registry = st.registry

    for plugin in PLUGINS:
        registry.register(plugin)



import abc
import asyncio
import dataclasses as dc
import enum
from datetime import datetime
from functools import reduce
from typing import Sequence, Coroutine, Dict, Optional, Any, Callable

import networkx as nx
import pluggy
from networkx.algorithms.dag import descendants

from statey import exc
from statey.hooks import hookspec, create_plugin_manager
from statey.resource import ResourceGraph
from statey.syms import utils, session
from statey.task import Task, TaskStatus, TaskInfo, ErrorInfo, CoroutineTask


class TaskGraph(abc.ABC):
    """
    Abstract base class for a task graph that a TaskGraphExecutor can execute
    """

    @abc.abstractmethod
    def set_status(
        self,
        key: str,
        status: TaskStatus,
        error: Optional[ErrorInfo] = None,
        skipped_by: Optional[str] = None,
    ) -> None:
        """
        Set the status of a given task
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_task(self, key: str) -> Task:
        """
        Get the task with the given key
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_info(self, key: str) -> TaskInfo:
        """
        Get the task with the given key
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_descendants(self, key: str) -> Sequence[str]:
        """
        Get all descendant tasks of `key`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_ready_tasks(self) -> Sequence[str]:
        """
        Get the tasks that are ready to schedule
        """
        raise NotImplementedError

    @abc.abstractmethod
    def keys(self) -> Sequence[str]:
        """
        Get all tasks in this graph
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class AsyncIOTaskGraph(TaskGraph):
    """
    Wrapper class for a flat task graph, offering API methods to manipulate it, usually with
    an executor
    """

    task_graph: nx.DiGraph

    def __post_init__(self) -> None:
        """
        Fill initial states for all tasks
        """
        for node in self.task_graph:
            self.task_graph.nodes[node]["info"] = TaskInfo(TaskStatus.NOT_STARTED)

    def set_status(
        self,
        key: str,
        status: TaskStatus,
        error: Optional[ErrorInfo] = None,
        skipped_by: Optional[str] = None,
    ) -> None:
        """
        Set the status of a given task
        """
        self.task_graph.nodes[key]["info"] = TaskInfo(
            status, error=error, skipped_by=skipped_by
        )

    def get_task(self, key: str) -> Task:
        """
        Get the task with the given key
        """
        if "task" not in self.task_graph.nodes[key]:
            coro_factory = self.task_graph.nodes[key]["coroutine_factory"]
            self.task_graph.nodes[key]["task"] = CoroutineTask(coro_factory)
        return self.task_graph.nodes[key]["task"]

    def get_info(self, key: str) -> TaskInfo:
        """
        Get the task with the given key
        """
        return self.task_graph.nodes[key]["info"]

    def get_descendants(self, key: str) -> Sequence[str]:
        """
        Get all descendant tasks of `key`
        """
        return descendants(self.task_graph, key)

    def get_ready_tasks(self) -> Sequence[str]:
        """
        Get the tasks that are ready to schedule
        """
        done_tasks = {
            node
            for node in self.task_graph.nodes
            if self.get_info(node).status.value > TaskStatus.SKIPPED.value
        }
        not_started_tasks = {
            node
            for node in self.task_graph.nodes
            if self.get_info(node).status == TaskStatus.NOT_STARTED
        }
        out = []
        for task_key in not_started_tasks:
            # We will mark tasks as ready if all their predecessors are done;
            # it's left to the executor to skip other tasks as needed
            if set(self.task_graph.pred[task_key]) - done_tasks:
                continue
            out.append(task_key)
        return out

    def keys(self) -> Sequence[str]:
        return list(self.task_graph.nodes)


@dc.dataclass(frozen=True)
class ResourceTaskGraph(TaskGraph):
    """
    Wrapper class for a flat task graph, offering API methods to manipulate it, usually with
    an executor
    """

    task_graph: nx.DiGraph
    output_session: session.Session
    resource_graph: ResourceGraph

    def __post_init__(self) -> None:
        """
        Fill initial states for all tasks
        """
        for node in self.task_graph:
            self.task_graph.nodes[node]["info"] = TaskInfo(TaskStatus.NOT_STARTED)

    def set_status(
        self,
        key: str,
        status: TaskStatus,
        error: Optional[ErrorInfo] = None,
        skipped_by: Optional[str] = None,
    ) -> None:
        self.task_graph.nodes[key]["info"] = TaskInfo(
            status, error=error, skipped_by=skipped_by
        )

    def get_task(self, key: str) -> Task:
        return self.task_graph.nodes[key]["task"]

    def get_info(self, key: str) -> TaskInfo:
        return self.task_graph.nodes[key]["info"]

    def get_descendants(self, key: str) -> Sequence[str]:
        return descendants(self.task_graph, key)

    def get_ready_tasks(self) -> Sequence[str]:
        done_tasks = {
            node
            for node in self.task_graph.nodes
            if self.get_info(node).status.value > TaskStatus.SKIPPED.value
        }
        not_started_tasks = {
            node
            for node in self.task_graph.nodes
            if self.get_info(node).status == TaskStatus.NOT_STARTED
        }
        out = []
        for task_key in not_started_tasks:
            # We will mark tasks as ready if all their predecessors are done;
            # it's left to the executor to skip other tasks as needed
            if set(self.task_graph.pred[task_key]) - done_tasks:
                continue
            out.append(task_key)
        return out

    def keys(self) -> Sequence[str]:
        return list(self.task_graph.nodes)


class ExecutionStrategy(enum.Enum):
    """
    Define different methods of execution a graph.
    EAGER means that we'll keep executing as long as we don't have a failure in a given
    branch.
    TENTATIVE means once we have one failure, we won't schedule any more tasks
    """

    EAGER = "eager"
    TENTATIVE = "tentative"


@dc.dataclass
class ExecutionInfo(utils.Cloneable):
    """
    Store some metadata about the result of an execution.
    """

    task_graph: TaskGraph
    strategy: ExecutionStrategy
    tasks: Dict[str, asyncio.Task] = dc.field(default_factory=dict)
    cancelled_by: Optional[str] = None
    cancelled_by_errors: Sequence[BaseException] = dc.field(default_factory=list)
    start_timestamp: datetime = dc.field(default_factory=datetime.utcnow)
    end_timestamp: datetime = dc.field(default=None)

    def is_success(self) -> bool:
        """
        Indicate whether this execution was a success
        """
        tasks_by_status = self.tasks_by_status()
        return set(tasks_by_status) <= {TaskStatus.SUCCESS}

    def tasks_by_status(self) -> Dict[TaskStatus, Sequence[str]]:
        """
        Sort tasks by status and return the names of tasks for each status
        """
        out = {}
        for key in self.task_graph.keys():
            info = self.task_graph.get_info(key)
            out.setdefault(info.status, []).append(key)
        return out

    def raise_for_failure(self) -> None:
        """
        Raise an error to indicate that this execution was not successful, if
        relevant.
        """
        if not self.is_success():
            raise exc.ExecutionError(self)


class TaskGraphHooks:
    """
    Specifies hooks to call during task graph execution
    """

    @hookspec
    def before_run(self, key: str, task: Task, executor: "TaskGraphExecutor") -> None:
        """
        Register side effects before a given task is run
        """

    @hookspec
    def after_run(
        self, key: str, task: Task, status: TaskStatus, executor: "TaskGraphExecutor"
    ) -> None:
        """
        Register side effects after a given task is run
        """

    @hookspec
    def caught_signal(
        self, signals: int, max_signals: int, executor: "TaskGraphExecutor"
    ) -> None:
        """
        Register side effects when a signal is caught, indicating whether the program
        is going to exit
        """

    @hookspec
    def task_wrapper(
        self, key: str, task: Task, executor: "TaskGraphExecutor"
    ) -> Callable[[Task], Task]:
        """
        Return a callable that wraps the given task with some additional behavior
        """


def create_executor_plugin_manager() -> pluggy.PluginManager:
    """
    Factory function for a plugin manager for a GraphExecutor
    """
    pm = create_plugin_manager()
    pm.add_hookspecs(TaskGraphHooks)
    return pm


class TaskGraphExecutor(abc.ABC):
    """
    Executes all the tasks in a given graph
    """

    pm: pluggy.PluginManager

    @abc.abstractmethod
    def execute(
        self,
        task_graph: TaskGraph,
        strategy: ExecutionStrategy = ExecutionStrategy.EAGER,
        max_signals: int = 2,
    ) -> ExecutionInfo:
        """
        Execute the tasks in the given graph, returning information about the run.
        """
        raise NotImplementedError


class AsyncTaskGraphExecutor(TaskGraphExecutor):
    """"""

    @abc.abstractmethod
    async def execute_async(
        self,
        task_graph: TaskGraph,
        exec_info: Optional[ExecutionInfo] = None,
        strategy: ExecutionStrategy = ExecutionStrategy.EAGER,
    ) -> ExecutionInfo:
        """
        Wrap all task executions into a single coroutine. This should _not_
        handle signals
        """
        raise NotImplementedError

    def hard_cancel(self, exec_info: ExecutionInfo) -> None:
        """
        Cancel any running tasks in exec_info
        """
        for task in exec_info.tasks.values():
            if not task.done():
                task.cancel()

    def task_loop(
        self, coro: Coroutine, max_signals: int, exec_info: ExecutionInfo
    ) -> Any:
        """
        Handle signals like KeyboardInterrupt and SystemExit properly, only cancelling running
        tasks when we reach max_signals.
        """
        signals = 0
        loop = asyncio.get_event_loop()
        coro_as_task = asyncio.ensure_future(coro)

        while True:
            try:
                loop.run_until_complete(coro_as_task)
                break
            except Exception:
                raise
            # Catch anything else (e.g. SystemExit, KeyboardInterrupt) and handle it
            # gracefully, avoiding cancelling running tasks as long as the user will
            # let us.
            except BaseException as err:
                signals += 1
                exec_info.cancelled_by_errors.append(err)

                will_cancel = signals >= max_signals

                self.pm.hook.caught_signal(
                    signals=signals, max_signals=max_signals, executor=self
                )

                if will_cancel:
                    self.hard_cancel(exec_info)

    def execute(
        self,
        task_graph: TaskGraph,
        strategy: ExecutionStrategy = ExecutionStrategy.EAGER,
        max_signals: int = 2,
    ) -> ExecutionInfo:

        exec_info = ExecutionInfo(task_graph, strategy)

        coro = self.execute_async(task_graph, exec_info, strategy)
        self.task_loop(coro, max_signals, exec_info)

        exec_info.end_timestamp = datetime.utcnow()

        return exec_info


class AsyncIOGraphExecutor(AsyncTaskGraphExecutor):
    """
    Graph executors that uses python asyncio couroutines for concurrency
    """

    def __init__(self, pm: Optional[pluggy.PluginManager] = None) -> None:
        if pm is None:
            pm = create_executor_plugin_manager()
        self.pm = pm

    async def after_task_success(self, key: str, exec_info: ExecutionInfo) -> None:
        """
        Callback after a task completes _successfully_
        """
        exec_info.task_graph.set_status(key, TaskStatus.SUCCESS)

        ready_tasks = exec_info.task_graph.get_ready_tasks()
        always_eager_coros = []
        other_coros = []
        for task in ready_tasks:
            coro = self.task_wrapper(
                task, exec_info.task_graph.get_task(task), exec_info
            )
            if exec_info.task_graph.get_task(task).always_eager():
                always_eager_coros.append(coro)
            else:
                other_coros.append(coro)

        # If there are any ready always eager tasks, we'll execute them right away no matter what
        if always_eager_coros:
            await asyncio.wait(always_eager_coros)

        # If the run has been cancelled and we're doing tentative execution, skip any unrun descendent
        # tasks
        if (
            exec_info.cancelled_by is not None
            and exec_info.strategy == ExecutionStrategy.TENTATIVE
        ):
            for child_key in exec_info.task_graph.get_descendants(key):
                if (
                    exec_info.task_graph.get_info(child_key).status
                    == TaskStatus.NOT_STARTED
                ):
                    exec_info.task_graph.set_status(
                        child_key, TaskStatus.SKIPPED, skipped_by=exec_info.cancelled_by
                    )
            return

        # If we get a signal to cancel, we'll stop creating new tasks regardless of the execution
        # strategy other than always eager tasks.
        if exec_info.cancelled_by_errors:
            for child_key in exec_info.task_graph.get_descendants(key):
                if (
                    exec_info.task_graph.get_info(child_key).status
                    == TaskStatus.NOT_STARTED
                ):
                    exec_info.task_graph.set_status(
                        child_key, TaskStatus.SKIPPED, skipped_by="ERROR"
                    )
            return

        # Otherwise, schehdule any non-always-eager tasks
        if other_coros:
            await asyncio.wait(other_coros)

    async def after_task_failure(
        self, key: str, exec_info: ExecutionInfo, error: ErrorInfo
    ) -> None:
        """
        Callback after a task fails
        """
        exec_info.task_graph.set_status(key, TaskStatus.FAILED, error=error)

        always_eager_coros = []

        ready_tasks = exec_info.task_graph.get_ready_tasks()

        for task_key in exec_info.task_graph.get_descendants(key):
            task = exec_info.task_graph.get_task(task_key)
            if task.always_eager():
                # If the task is always eager we won't skip it, and if it's ready to
                # run we will actually trigger it here. If this is something like a graph
                # operation it's alright for it to fail, that simply means the graph
                # won't be mutated. It supports partial checkpoints for tasks
                if task_key in ready_tasks:
                    coro = self.task_wrapper(task_key, task, exec_info)
                    always_eager_coros.append(coro)
            else:
                exec_info.task_graph.set_status(
                    task_key, TaskStatus.SKIPPED, skipped_by=key
                )

        if exec_info.cancelled_by is None:
            exec_info.cancelled_by = key

        if always_eager_coros:
            await asyncio.wait(always_eager_coros)

    async def task_wrapper(
        self, key: str, task: Task, exec_info: ExecutionInfo
    ) -> None:
        """
        Wrap a coroutine to catch exceptions and set the appropriate result
        """
        # This task may have been scheduled multiple times--if so, just exit here
        if exec_info.task_graph.get_info(key).status != TaskStatus.NOT_STARTED:
            return

        self.pm.hook.before_run(key=key, task=task, executor=self)

        wrappers = self.pm.hook.task_wrapper(key=key, task=task, executor=self)
        wrappers = [wrapper for wrapper in wrappers if wrapper is not None]
        wrapped_task = reduce(lambda x, y: y(x), wrappers, task)

        asyncio_task = asyncio.ensure_future(wrapped_task.run())
        exec_info.tasks[key] = asyncio_task
        exec_info.task_graph.set_status(key, TaskStatus.PENDING)
        try:
            await asyncio_task
        except Exception:
            self.pm.hook.after_run(
                key=key, task=task, status=TaskStatus.FAILED, executor=self
            )
            await self.after_task_failure(key, exec_info, ErrorInfo.exc_info())
        else:
            self.pm.hook.after_run(
                key=key, task=task, status=TaskStatus.SUCCESS, executor=self
            )
            await self.after_task_success(key, exec_info)

    async def execute_async(
        self,
        task_graph: TaskGraph,
        exec_info: Optional[ExecutionInfo] = None,
        strategy: ExecutionStrategy = ExecutionStrategy.EAGER,
    ) -> None:

        if exec_info is None:
            exec_info = ExecutionInfo(task_graph, strategy)

        ready_tasks = {
            key: task_graph.get_task(key) for key in task_graph.get_ready_tasks()
        }
        ready_wrappers = [
            self.task_wrapper(key, task, exec_info) for key, task in ready_tasks.items()
        ]

        if not ready_wrappers:
            return exec_info

        await asyncio.wait(ready_wrappers)

        exec_info.end_timestamp = datetime.utcnow()

        return exec_info



import copy
import dataclasses as dc
import sys
import textwrap as tw
from itertools import chain, product
from typing import Any, Union, Iterator, Sequence, Callable

import networkx as nx
from networkx.algorithms.dag import (
    is_directed_acyclic_graph,
    topological_sort,
)

import statey as st
from statey import exc
from statey.syms import types, utils, session, impl, Object, stack


class PythonNamespace(session.Namespace):
    """
    Pure python namespace implementation
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.types = {}

    def new(self, key: str, type: types.Type) -> Object:
        """
        Create a new symbol for the given key and schema and add it to the current namespace
        """
        self.path_parser.validate_name(key)
        if key in self.types:
            raise exc.DuplicateSymbolKey(key, self)
        self.types[key] = type
        ref = self.ref(key)
        return Object(ref, frame=stack.frame_snapshot(1))

    def delete(self, key: str) -> None:
        self.resolve(key)
        del self.types[key]

    def keys(self) -> Iterator[str]:
        return self.types.keys()

    def resolve(self, key: str) -> types.Type:
        """
        Get the type of the given key, raising an error if it is not in the current schema
        """
        base, *rel_path = self.path_parser.split(key)
        if base not in self.types:
            raise exc.SymbolKeyError(key, self)

        base_semantics = self.registry.get_semantics(self.types[base])
        try:
            semantics = base_semantics.path_semantics(rel_path)
        except AttributeError as err:
            raise exc.SymbolKeyError(key, self) from err

        return semantics.type

    def clone(self) -> session.Namespace:
        new_inst = PythonNamespace(self.registry, self.path_parser)
        new_inst.types = self.types.copy()
        return new_inst


@dc.dataclass
class ResolutionStack:
    """
    A resolution stack object contains logic to format stack information
    for resolution errors
    """

    dag: nx.DiGraph
    resolving_obj_id: Any
    current_obj_stack: Sequence[Any] = ()

    def __post_init__(self) -> None:
        self.current_obj_stack = list(self.current_obj_stack)

    def get_object(self, symbol_id: Any) -> Object:
        return self.dag.nodes[symbol_id]["symbol"]

    def get_object_stack(self) -> Sequence[stack.FrameSnapshot]:
        """
        Get a list of FrameSnapshots for objects between the resolving object and the error object
        in the DAG.
        """
        if not self.current_obj_stack:
            obj = self.get_object()
            return [(obj._frame, [obj])]

        frame_path = []
        objects = []
        last_frame = None

        full_path = list(
            reversed(
                nx.shortest_path(
                    self.dag, self.current_obj_stack[0], self.resolving_obj_id
                )
            )
        )
        if not full_path:
            full_path.append(self.current_obj_stack[0])
        else:
            full_path.extend(self.current_obj_stack[1:])

        for obj_id in full_path:
            obj = self.get_object(obj_id)
            frame = obj._frame
            if frame != last_frame:
                if objects:
                    frame_path.append((last_frame, objects))
                objects = []

            objects.append(obj)
            last_frame = frame

        if last_frame and objects:
            frame_path.append((last_frame, objects))

        return frame_path

    def push(self, obj_id: Any) -> None:
        """"""
        self.current_obj_stack.append(obj_id)

    def pop(self) -> None:
        """"""
        self.current_obj_stack.pop()


class PythonSession(session.Session):
    """
    A pure python session implementation for resolving objects.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.data = {}

    def set_data(self, key: str, data: Any) -> None:
        """
        Set the given data for the given key.
        """
        typ = self.ns.resolve(key)

        encoder = self.ns.registry.get_encoder(typ, serializable=True)
        try:
            encoded_data = encoder.encode(data)

            root, *rel_path = self.ns.path_parser.split(key)
            if rel_path:
                raise ValueError("Data can only be set on the root level.")
        except Exception as err:
            raise exc.InvalidDataError(key, err) from err

        self.data[key] = encoded_data

    def delete_data(self, key: str) -> None:
        if key not in self.data:
            raise exc.SymbolKeyError(key, self.ns)
        del self.data[key]

    def get_encoded_data(self, key: str) -> Any:
        """
        Return the data or symbol at the given key, with a boolean is_symbol also returned indicating
        whether that key points to a symbol. If the key has not been set, syms.exc.MissingDataError
        will be raised
        """
        typ = self.ns.resolve(key)
        base, *rel_path = self.ns.path_parser.split(key)
        if base not in self.data:
            raise exc.MissingDataError(key, typ, self)

        base_type = self.ns.resolve(base)
        base_semantics = self.ns.registry.get_semantics(base_type)
        value = self.data[base]
        for idx, attr in enumerate(rel_path):
            try:
                value = base_semantics.get_attr(value, attr)
            except (KeyError, AttributeError) as err:
                sub_path = [base] + list(rel_path)[: idx + 1]
                sub_key = self.ns.path_parser.join(sub_path)
                raise exc.MissingDataError(key, base_type, self) from err
            finally:
                base_semantics = base_semantics.attr_semantics(attr)

        return value

    @stack.resolutionstart
    def _build_symbol_dag(
        self,
        symbol: Object,
        graph: nx.DiGraph,
        resolution_stack: ResolutionStack,
        check_dag: bool = True,
    ) -> nx.DiGraph:

        syms = [(symbol, ())]
        added = set()

        while syms:
            sym, downstreams = syms.pop(0)
            resolution_stack.push(sym._impl.id)
            try:
                processed = False
                if sym._impl.id in graph.nodes:
                    processed = True
                else:
                    graph.add_node(sym._impl.id, symbol=sym)
                    added.add(sym._impl.id)

                for symbol_id in downstreams:
                    if (sym._impl.id, symbol_id) not in graph.edges:
                        graph.add_edge(sym._impl.id, symbol_id)
                        added.add(symbol_id)
                        added.add(sym._impl.id)

                if processed:
                    continue

                for upstream in sym._inst.depends_on(self):
                    syms.append((upstream, (sym._impl.id,)))
            except Exception as err:
                _, _, tb = sys.exc_info()
                raise exc.ResolutionError(resolution_stack, err, tb) from err
            except BaseException:
                raise
            else:
                resolution_stack.pop()

        if check_dag and added:
            # Check that this is in fact a DAG and has no cycles
            self._raise_not_dag(graph, list(added))

        return added

    def _raise_not_dag(self, graph: nx.DiGraph, sources=None) -> None:
        try:
            nx.find_cycle(graph, sources)
        except nx.NetworkXNoCycle:
            pass
        else:
            raise ValueError(f"Graph is not a DAG!")

    @stack.internalcode
    def _process_symbol_dag(
        self,
        obj_id: Any,
        dag: nx.DiGraph,
        resolution_stack: ResolutionStack,
        allow_unknowns: bool = False,
    ) -> None:

        # @stack.internalcode
        @stack.resolutionstart
        def handle_symbol_id(symbol_id):

            sym = dag.nodes[symbol_id]["symbol"]
            if "result" in dag.nodes[symbol_id]:
                return dag.nodes[symbol_id]["result"]

            try:
                result = sym._inst.apply(dag, self)
            except exc.UnknownError as err:
                if not allow_unknowns:
                    raise
                if err.expected is utils.MISSING:
                    result = Object(impl.Unknown(sym))
                else:
                    expected = Object(err.expected, sym._type, sym._registry)
                    result = self.resolve(expected, decode=False, allow_unknowns=True)
            else:
                semantics = self.ns.registry.get_semantics(sym._type)
                expanded_result = semantics.expand(result)

                def resolve_child(x):
                    added = self._build_symbol_dag(
                        x, dag, resolution_stack, check_dag=False
                    )
                    if (x._impl.id, symbol_id) not in dag.edges:
                        dag.add_edge(x._impl.id, symbol_id)
                        self._raise_not_dag(dag, list(added | {x._impl.id}))
                    elif added:
                        self._raise_not_dag(dag, list(added))

                    ancestors = set(dag.pred[x._impl.id])
                    for node in list(ancestors):
                        if "result" in dag.nodes[node]:
                            ancestors.remove(node)

                    for sym_id in list(topological_sort(dag.subgraph(ancestors))):
                        wrapped_handle_symbol_id(sym_id)

                    return wrapped_handle_symbol_id(x._impl.id)

                result = semantics.map_objects(resolve_child, expanded_result)

            dag.nodes[symbol_id]["result"] = result
            return result

        @stack.internalcode
        def wrapped_handle_symbol_id(symbol_id):
            exception, tb = None, None
            resolution_stack.push(symbol_id)
            try:
                return handle_symbol_id(symbol_id)
            except Exception as err:
                _, _, tb = sys.exc_info()
                exception = exc.ResolutionError(resolution_stack, err, tb)
            except BaseException:
                raise
            else:
                resolution_stack.pop()
            raise exception.with_traceback(tb)

        ancestors = set(nx.ancestors(dag, obj_id)) | {obj_id}
        for symbol_id in list(topological_sort(dag)):
            if symbol_id not in ancestors:
                continue
            wrapped_handle_symbol_id(symbol_id)

    def _digraph(self) -> nx.DiGraph:
        return nx.DiGraph()

    def resolve(
        self, symbol: Object, allow_unknowns: bool = False, decode: bool = True
    ) -> Any:

        graph = self._digraph()

        resolution_stack = ResolutionStack(graph, symbol._impl.id)
        with stack.rewrite_ctx(resolution_stack):
            self._build_symbol_dag(symbol, graph, resolution_stack)
            self._process_symbol_dag(
                symbol._impl.id, graph, resolution_stack, allow_unknowns
            )

        encoded = graph.nodes[symbol._impl.id]["result"]
        if not decode:
            return encoded

        encoder = self.ns.registry.get_encoder(symbol._type)
        return encoder.decode(encoded)

    def dependency_graph(self) -> nx.MultiDiGraph:
        graph = nx.MultiDiGraph()

        dag = self._digraph()

        key = None
        obj_ids = set()
        for key in self.ns.keys():
            ref = self.ns.ref(key)
            resolution_stack = ResolutionStack(dag, ref._impl.id)
            with stack.rewrite_ctx(resolution_stack):
                self._build_symbol_dag(ref, dag, resolution_stack, check_dag=False)
                self._process_symbol_dag(
                    ref._impl.id, dag, resolution_stack, allow_unknowns=True
                )

            obj_ids.add(ref._impl.id)

        self._raise_not_dag(graph, list(obj_ids))

        # Make a copy of `dag` since it will be mutated
        dag = dag.copy()

        ref_symbol_ids = set()

        for node in dag.nodes:
            if isinstance(dag.nodes[node]["symbol"]._impl, impl.Reference):
                ref_symbol_ids.add(node)

        utils.subgraph_retaining_dependencies(dag, ref_symbol_ids)

        for node in list(nx.topological_sort(dag)):
            ref = dag.nodes[node]["symbol"]

            base, *rel_path = self.ns.path_parser.split(ref._impl.path)
            graph.add_node(base)

            for pred in dag.pred[node]:
                pref_ref = dag.nodes[pred]["symbol"]
                pred_base, *pred_rel_path = self.ns.path_parser.split(
                    pref_ref._impl.path
                )

                graph.add_edge(
                    pred_base, base, from_path=pred_rel_path, to_path=rel_path
                )

        return graph

    def clone(self) -> session.Session:
        new_inst = PythonSession(self.ns.clone())
        # Since data can only be set at the root level and is immutable while in the
        # session, a shallow copy works fine here.
        new_inst.data = self.data.copy()
        for plugin in self.pm.get_plugins():
            if plugin is not self:
                new_inst.pm.register(plugin)
        return new_inst


class CachingPythonSession(PythonSession):
    """
    PythonSession that holds a single graph per instance
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.graph = nx.DiGraph()

    def clone(self) -> session.Session:
        new_inst = super().clone()
        new_inst.graph = self.graph.copy()
        return new_inst

    def _invalidate_symbol(self, obj_id: Any) -> None:
        if obj_id not in self.graph.nodes:
            return

        for node in list(nx.descendants(self.graph, obj_id)):
            if "result" in self.graph.nodes[node]:
                del self.graph.nodes["result"]
            # self.graph.remove_node(node)

        self.graph.remove_node(obj_id)

    def _clean_unknowns(self) -> None:
        for node in list(self.graph.nodes):
            if node not in self.graph.nodes:
                continue
            data = self.graph.nodes[node]
            if "result" in data and isinstance(data["result"], st.Object):
                self._invalidate_symbol(node)

    def resolve(
        self, symbol: Object, allow_unknowns: bool = False, decode: bool = True
    ) -> Any:
        res = super().resolve(symbol, allow_unknowns, decode)
        if allow_unknowns:
            self._clean_unknowns()
        return res

    def set_data(self, key: str, data: Any) -> None:
        invalidate = False
        if key in self.data:
            try:
                current = self.resolve(self.ns.ref(key))
            except st.exc.ResolutionError:
                invalidate = True

            if not invalidate and isinstance(data, Object):
                try:
                    new_value = self.resolve(data)
                except st.exc.ResolutionError:
                    invalidate = True
                else:
                    invalidate = new_value != current

        if invalidate:
            self._invalidate_symbol(self.ns.ref(key)._impl.id)

        super().set_data(key, data)

    def delete_data(self, key: str) -> None:
        super().delete_data(key, data)
        self._invalidate_key(key)


def create_session() -> session.Session:
    """
    Default factory for creating the best session given the runtime
    """
    return CachingPythonSession(PythonNamespace())



import abc
import asyncio
import dataclasses as dc
from collections import Counter
from typing import (
    Optional,
    Dict,
    Any,
    Sequence,
    Type as PyType,
    Tuple,
    Iterator,
)

import marshmallow as ma
import networkx as nx

import statey as st
from statey import exc
from statey.helpers import async_providers_context
from statey.provider import ProviderId, Provider, ProviderIdSchema
from statey.task import TaskSession
from statey.syms import types, utils, session, impl, Object


class StateSchema(ma.Schema):
    """
    Marshmallow schema for extra safety encoding/decoding states
    """

    name = ma.fields.Str(required=True, default=None)
    input_type = ma.fields.Dict(required=True, default=None)
    output_type = ma.fields.Dict(required=True, default=None)
    null = ma.fields.Bool(required=True, default=None)


class AbstractState(abc.ABC):
    """
    Abstract base class defining required attributes for states
    """

    name: str
    input_type: types.Type
    output_type: types.Type
    null: bool

    @property
    def type(self) -> types.Type:
        return self.output_type

    def to_dict(self, registry: "Registry") -> Dict[str, Any]:
        """
        Render this state to a JSON-serializable dictionary
        """
        input_type_serializer = registry.get_type_serializer(self.input_type)
        output_type_serializer = registry.get_type_serializer(self.output_type)
        out = {
            "name": self.name,
            "input_type": input_type_serializer.serialize(self.input_type),
            "output_type": output_type_serializer.serialize(self.output_type),
            "null": self.null,
        }
        return StateSchema().dump(out)

    @classmethod
    def from_dict(cls, data: Dict[str, Any], registry: "Registry") -> "State":
        """
        Render a State from the output of to_dict()
        """
        data = StateSchema().load(data)
        input_type_serializer = registry.get_type_serializer_from_data(
            data["input_type"]
        )
        input_type = input_type_serializer.deserialize(data["input_type"])
        output_type_serializer = registry.get_type_serializer_from_data(
            data["output_type"]
        )
        output_type = output_type_serializer.deserialize(data["output_type"])
        return cls(
            name=data["name"],
            input_type=input_type,
            output_type=output_type,
            null=data["null"],
        )


@dc.dataclass(frozen=True)
class State(AbstractState):
    """
    A state corresponds to some type for a resource.
    """

    name: str
    input_type: types.Type
    output_type: types.Type
    null: bool = dc.field(repr=False, default=False)


@dc.dataclass(frozen=True)
class NullState(State):
    """
    Null states must always have types.EmptyType as their type, so
    this is a helper to create such states.
    """

    input_type: types.Type = dc.field(init=False, default=types.EmptyType)
    output_type: types.Type = dc.field(init=False, default=types.EmptyType)
    null: bool = dc.field(repr=False, init=False, default=True)


class ResourceStateSchema(ma.Schema):
    """
    Marshmallow schema for extra safety encoding/decoding resource states
    """

    state = ma.fields.Nested(StateSchema(), required=True, default=None)
    resource = ma.fields.Str(required=True, default=None)
    provider = ma.fields.Nested(ProviderIdSchema(), required=True, default=None)


@dc.dataclass(frozen=True)
class ResourceState:
    """
    A resource state is a state that is bound to some resource.
    """

    state: AbstractState
    resource: str
    provider: ProviderId

    @property
    def name(self) -> str:
        return self.state.name

    @property
    def input_type(self) -> types.Type:
        return self.state.input_type

    @property
    def output_type(self) -> types.Type:
        return self.state.output_type

    @property
    def type(self) -> types.Type:
        return self.state.type

    @property
    def null(self) -> bool:
        return self.state.null

    def snapshot(self, data: Any) -> "StateSnapshot":
        """
        Create a StateSnapshot from this resource state and the given data
        """
        return StateSnapshot(data, self)

    def __call__(self, arg=utils.MISSING, **kwargs) -> "BoundState":
        """
        Factory method for creating bound states for this resource state.
        """
        if arg is utils.MISSING:
            arg = kwargs
        elif kwargs:
            raise ValueError(
                "Either one positional arg or keyword arguments are required for a BoundState."
            )
        return BoundState(self, arg)

    def to_dict(self, registry: "Registry") -> Dict[str, Any]:
        """
        Render this resource state to a JSON-serializable dictionary
        """
        out = {
            "resource": self.resource,
            "state": self.state.to_dict(registry),
            "provider": self.provider.to_dict(),
        }
        return ResourceStateSchema().dump(out)

    @classmethod
    def from_dict(cls, data: Dict[str, Any], registry: "Registry") -> "ResourceState":
        """
        Render a ResourceState from the output of to_dict()
        """
        data = ResourceStateSchema().load(data)
        state = State.from_dict(data["state"], registry)
        provider_id = ProviderId.from_dict(data["provider"])
        return cls(resource=data["resource"], state=state, provider=provider_id)


@dc.dataclass(frozen=True)
class BoundState(utils.Cloneable):
    """
    Describes an input configuration for a resource
    """

    state: ResourceState
    input: "Object"
    output: "Object" = dc.field(init=False, default=None)

    def __post_init__(self) -> None:
        input_obj = st.Object(self.input, self.state.input_type)
        self.__dict__["input"] = input_obj
        output_impl = impl.Unknown(
            refs=(input_obj,), return_type=self.state.output_type
        )
        self.__dict__["output"] = st.Object(output_impl)

    def bind(self, registry: "Registry") -> None:
        """
        Bind input and output objects to the given registry
        """
        self.__dict__["input"] = st.Object(self.input, registry=registry)
        self.__dict__["output"] = st.Object(self.output, registry=registry)


class StateTuple(abc.ABC, utils.Cloneable):
    """
    Binding a state to a value
    """

    obj: "Object"
    state: ResourceState

    @property
    @abc.abstractmethod
    def type(self) -> types.Type:
        """
        Return the type of `data`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_obj(self, registry: Optional["Registry"] = None) -> "Object":
        """
        Obtain an object representation of this StateTuple
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class StateConfig(StateTuple):
    """
    A state configuration object, may contain unknowns
    """

    obj: "Object"
    state: ResourceState

    @property
    def type(self) -> types.Type:
        return self.state.input_type

    def get_obj(self, registry: Optional["Registry"] = None) -> "Object":
        if registry is None:
            registry = st.registry
        return st.Object(self.obj._impl, self.obj._type, registry)


@dc.dataclass(frozen=True)
class StateSnapshot(StateTuple):
    """
    A fully-resolved snapshot of a resource state
    """

    data: Any
    state: ResourceState

    @property
    def type(self) -> types.Type:
        return self.state.output_type

    def get_obj(self, registry: Optional["Registry"] = None) -> "Object":
        if registry is None:
            registry = st.registry
        return st.Object(self.data, self.type, registry)

    @property
    def obj(self) -> "Object":
        return self.get_obj()


class Resource(abc.ABC):
    """
    A resource represents a stateful object of some kind, and it can have one
    or more "states" that that object can exist in.
    """

    name: str
    provider: Provider

    @property
    @abc.abstractmethod
    def null_state(self) -> ResourceState:
        """
        Return the null state of this resource, at which point it will removed from
        resource graphs
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def plan(
        self, current: StateSnapshot, config: StateConfig, session: TaskSession
    ) -> Object:
        """
        Given a task session, the current state of a resource, and a task session with
        corresponding input reference, return an output reference that can be fully
        resolved when all the tasks in the task session have been complete successfully.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def refresh(self, current: StateSnapshot) -> StateSnapshot:
        """
        Given the current bound state, return a new bound state that represents that actual
        current state of the object
        """
        raise NotImplementedError

    async def finalize(self, data: StateSnapshot) -> StateSnapshot:
        """
        Called on states before they are committed to the resource graph. By default a noop
        """
        return data


class ResourceSession(session.Session):
    """
    Session subclass that wraps a regular session but handles resources in a special manner.
    """

    def __init__(self, session: session.Session) -> None:
        super().__init__(session.ns)
        self.session = session
        self.states = {}
        self.pm.register(self)

    @st.hookimpl
    def before_set(self, key: str, value: Any) -> Tuple[Any, types.Type]:
        if not isinstance(value, BoundState):
            return
        # Bind the config to this registry
        value.bind(self.ns.registry)
        self.states[key] = value
        return value.output, value.output._type

    def resolve(
        self, symbol: Object, allow_unknowns: bool = False, decode: bool = True
    ) -> Any:
        return self.session.resolve(symbol, allow_unknowns, decode)

    def set_data(self, key: str, data: Any) -> None:
        return self.session.set_data(key, data)

    def delete_data(self, key: str) -> None:
        return self.session.delete_data(key)

    def dependency_graph(self) -> nx.MultiDiGraph:
        return self.session.dependency_graph()

    def resource_graph(self) -> "ResourceGraph":
        """
        Return a fully-rendered ResourceGraph for this session. Will fail with a resolution
        error if data is missing or there are unknowns or unresolved futures in the session.
        """
        graph = ResourceGraph()
        dep_graph = self.dependency_graph()

        for key in self.ns.keys():
            if key in self.states:
                state = self.states[key]
                resolved = self.resolve(self.ns.ref(key), decode=False)
                graph.set(
                    key=key,
                    value=resolved,
                    type=state.state.output_type,
                    state=state.state,
                )
            else:
                resolved = self.resolve(self.ns.ref(key), decode=False)
                graph.set(key=key, value=resolved, type=self.ns.resolve(key))

        for node in dep_graph:
            if dep_graph.pred[node]:
                graph.add_dependencies(node, list(dep_graph.pred[node]))

        return graph

    def get_state(self, name: str) -> BoundState:
        """
        Get the resource state for the given name, raising SymbolKeyError to indicate
        a failure
        """
        if name not in self.states:
            raise exc.SymbolKeyError(name, self)
        return self.states[name]

    def clone(self) -> "ResourceSession":
        new_inst = ResourceSession(self.session.clone())
        new_inst.states = self.states.copy()
        for plugin in self.pm.get_plugins():
            if plugin is not self:
                new_inst.pm.register(plugin)
        return new_inst


def create_resource_session(
    session: Optional[session.Session] = None,
) -> ResourceSession:
    """
    Convenient factory function for creating resource sessions.
    """
    if session is None:
        import statey as st

        session = st.create_session()
    return ResourceSession(session)


@dc.dataclass(frozen=True)
class ResourceGraph:
    """
    A resource graph is a wrapper around a simple graph that stores similar information
    to a resource session, but without objects. Resource graphs are serializable. Note that
    not every node in a resource graph is necessarily a resource, it can contain any name
    just like a session.
    """

    graph: nx.DiGraph = dc.field(default_factory=nx.DiGraph)

    def set(
        self,
        key: str,
        value: Any,
        type: types.Type,
        state: Optional[ResourceState] = None,
        remove_dependencies: bool = True,
    ) -> None:
        """
        Set the given value in the graph, with a state optionally specified for resources. Note
        that this operation will remove the current upstream edges of `key` unless
        remove_upstreams=False is specified as an argument
        """
        if remove_dependencies and key in self.graph.nodes:
            for pred in list(self.graph.pred[key]):
                self.graph.remove_edge(pred, key)

        self.graph.add_node(key, value=value, type=type, state=state)

    def delete(self, key: str) -> None:
        """
        Delete the given key from this graph. Will raise KeyError if it doesn't exist
        """
        if key not in self.graph.nodes:
            raise KeyError(key)
        self.graph.remove_node(key)

    def add_dependencies(self, key: str, dependencies: Sequence[str]) -> None:
        """
        Add dependencies to the given key
        """
        if key not in self.graph.nodes:
            raise KeyError(key)
        for dep in dependencies:
            if dep not in self.graph.nodes:
                raise KeyError(dep)

        self.graph.add_edges_from([(dep, key) for dep in dependencies])

    async def refresh(
        self, registry: st.Registry, finalize: bool = False
    ) -> Iterator[str]:
        """
        Refresh the current state of all resources in the graph. Returns an asynchronous
        generator that yields keys as they finish refreshing successfully.
        """

        providers = {}

        for node in self.graph.nodes:
            state = self.graph.nodes[node]["state"]
            if state is None:
                continue
            provider = providers.get(state.provider)
            if provider is None:
                provider = providers[state.provider] = registry.get_provider(
                    *state.provider
                )

        async def handle_node(key):
            data = self.graph.nodes[key]
            state = data["state"]
            if state is None:
                return
            provider = providers[state.provider]
            resource = provider.get_resource(state.resource)
            result = await resource.refresh(StateSnapshot(data["value"], state))
            if finalize:
                result = await resource.finalize(result)
            self.set(
                key=key,
                value=result.data,
                type=result.type,
                state=result.state,
                remove_dependencies=False,
            )

        providers_list = list(providers.values())
        node_list = list(self.graph.nodes)
        coros = list(map(handle_node, node_list))

        async with async_providers_context(providers_list):
            for key, coro in zip(node_list, asyncio.as_completed(coros)):
                await coro
                yield key

    def clone(self) -> "ResourceGraph":
        return type(self)(self.graph.copy())

    def to_dict(self, registry: "Registry") -> Dict[str, Any]:
        """
        Return a JSON-serializable dictionary representation of this resource graph.
        """
        nodes = {}

        for node in self.graph.nodes:
            data = self.graph.nodes[node]

            type_serializer = registry.get_type_serializer(data["type"])
            state_dict = None
            if data["state"]:
                state_dict = data["state"].to_dict(registry)
                del state_dict["state"]["output_type"]

            nodes[node] = {
                "data": data["value"],
                "type": type_serializer.serialize(data["type"]),
                "state": state_dict,
                "depends_on": list(self.graph.pred[node]),
            }

        return nodes

    @classmethod
    def from_dict(cls, data: Any, registry: "Registry") -> "ResourceGraph":
        """
        Render a ResourceGraph from a JSON-serializable representation
        """
        deps = {}
        instance = cls()

        for key, node in data.items():
            if node["state"] is not None:
                type_dict = node["type"]
                state_dict_copy = node["state"].copy()
                state_dict_copy["state"] = state_dict_copy["state"].copy()
                state_dict_copy["state"]["output_type"] = type_dict
                state = ResourceState.from_dict(state_dict_copy, registry)
                typ = state.state.type
                instance.set(key, node["data"], typ, state)
            else:
                type_serializer = registry.get_type_serializer_from_data(node["type"])
                typ = type_serializer.deserialize(node["type"])
                instance.set(key, node["data"], typ)

            deps[key] = node["depends_on"]

        for to_node, from_nodes in deps.items():
            instance.add_dependencies(to_node, from_nodes)

        return instance



"""
Common functions for use in graph generation
"""
import itertools
import sys
from functools import partial
from typing import Optional, Dict, Any

try:
    from fmt.fmt import Parser, Expression, Constant, generate
except ImportError as err:
    raise RuntimeError(
        "`fmt` must be installed to use the statey.syms.fmt module; this "
        "can be done by including the `fmt` extra in your statey install i.e. "
        "statey[fmt]"
    )


import statey as st
from statey.syms.object_ import Object


def _eval_name(name: str, expr: str, namespace: Dict[str, Any]):
    new_ns = {}
    try:
        # pylint: disable=exec-used
        exec(f"{name}={expr}", namespace.copy(), new_ns)
    except NameError as exc:
        name = exc.args[0].split("'", 3)[1]
        raise NameError(f'"{name}" cannot be found in any namespaces.') from exc
    return new_ns[name]


def f(template: str, scope: Optional[Dict[str, Any]] = None) -> Object:
    """
    Interpolation operator for objects
    """
    variables = {} if scope is None else scope.copy()
    frame = sys._getframe(1)  # pylint: disable=protected-access
    variables.update(frame.f_globals)
    variables.update(frame.f_locals)

    nodes = Parser(template).parse()

    _next = partial(next, itertools.count(1))
    next_param = lambda: f"param_{_next()}"

    out_scope = {}
    out_nodes = []

    for node in nodes:
        if isinstance(node, Expression):
            # pylint: disable=protected-access,no-member
            obj = _eval_name(node._name, node._expr, variables)
        elif isinstance(node, Constant):
            # pylint: disable=protected-access,no-member
            obj = _eval_name(node._name, node._name, variables)
        else:
            out_nodes.append(node)
            continue

        param_name = next_param()  # pylint: disable=protected-access
        out_scope[param_name] = obj
        out_nodes.append(Constant(param_name, "{" + param_name + "}"))

    @st.function
    def render_nodes(scope: Dict[str, Any]) -> str:
        return generate(out_nodes, scope)

    return render_nodes(st.Object(out_scope, Dict[str, Any]))



import abc
import dataclasses as dc
from typing import Any, Type as PyType, Dict, Optional

import statey as st
from statey.syms import types


class TypeSerializer(abc.ABC):
    """
    A type serializer encodoes/decodes a type to/from a JSON-serializable value
    """

    @abc.abstractmethod
    def serialize(self, type: types.Type) -> Any:
        """
        Serialize the given type to a dictionary or some JSONI-serializable value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(self, data: Any) -> types.Type:
        """
        Deserialize the gien data (which should be the output of a serialize() call)
        into a native types.Type
        """
        raise NotImplementedError


class ValueTypeSerializer(TypeSerializer):
    """
    This serializes simple "value" types like ints, booleans
    """

    type_cls: PyType[types.Type]
    type_name: str

    def serialize(self, type: types.Type) -> Any:
        result = {"type": self.type_name}
        if type.nullable:
            result["nullable"] = True
        return result

    def deserialize(self, data: Any) -> types.Type:
        return self.type_cls(nullable=data.get("nullable", False))

    @classmethod
    @st.hookimpl
    def get_type_serializer(
        cls, type: types.Type, registry: "Registry"
    ) -> "TypeSerializer":
        if not isinstance(type, cls.type_cls):
            return None
        return cls()

    @classmethod
    @st.hookimpl
    def get_type_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> "TypeSerializer":
        if data.get("type") != cls.type_name:
            return None
        return cls()


@dc.dataclass(frozen=True)
class IntegerTypeSerializer(ValueTypeSerializer):
    """
    Value type serializer for ints
    """

    type_cls = types.IntegerType
    type_name = "integer"


@dc.dataclass(frozen=True)
class FloatTypeSerializer(ValueTypeSerializer):
    """
    Value type serializer for floats
    """

    type_cls = types.FloatType
    type_name = "number"


@dc.dataclass(frozen=True)
class BooleanTypeSerializer(ValueTypeSerializer):
    """
    Value type serializer for booleans
    """

    type_cls = types.BooleanType
    type_name = "boolean"


@dc.dataclass(frozen=True)
class StringTypeSerializer(ValueTypeSerializer):
    """
    Value type serializer for strings
    """

    type_cls = types.StringType
    type_name = "string"


@dc.dataclass(frozen=True)
class AnyTypeSerializer(ValueTypeSerializer):
    """
    Value type serializer for Any
    """

    type_cls = types.AnyType
    type_name = "any"


@dc.dataclass(frozen=True)
class ArrayTypeSerializer(TypeSerializer):
    """
    Type serializer for arrays
    """

    element_serializer: "TypeSerializer"

    def serialize(self, type: types.Type) -> Any:
        element_type = self.element_serializer.serialize(type.element_type)
        result = {"type": "array", "items": element_type}
        if type.nullable:
            result["nullable"] = True
        return result

    def deserialize(self, data: Any) -> types.Type:
        element_type = self.element_serializer.deserialize(data["items"])
        return types.ArrayType(
            element_type=element_type, nullable=data.get("nullable", False)
        )

    @classmethod
    @st.hookimpl
    def get_type_serializer(
        cls, type: types.Type, registry: "Registry"
    ) -> "TypeSerializer":
        if not isinstance(type, types.ArrayType):
            return None
        element_serializer = registry.get_type_serializer(type.element_type)
        return cls(element_serializer)

    @classmethod
    @st.hookimpl
    def get_type_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> "TypeSerializer":
        if data.get("type") != "array":
            return None
        element_type_data = data["items"]
        element_serializer = registry.get_type_serializer_from_data(element_type_data)
        return cls(element_serializer)


@dc.dataclass(frozen=True)
class MapTypeSerializer(TypeSerializer):
    """
    Type serializer for arrays
    """

    key_serializer: TypeSerializer
    value_serializer: TypeSerializer

    def serialize(self, type: types.Type) -> Any:
        key_type = self.key_serializer.serialize(type.key_type)
        value_type = self.value_serializer.serialize(type.value_type)
        result = {
            "type": "object",
            "additionalProperties": value_type,
        }

        # This is assumed to always be the case in JSON schemas, so omit this
        # information if it's just a simple string
        if key_type != {"type": "string"}:
            result["keys"] = key_type

        if type.nullable:
            result["nullable"] = True
        return result

    def deserialize(self, data: Any) -> types.Type:
        key_type = self.key_serializer.deserialize(data.get("keys", {"type": "string"}))
        value_type = self.value_serializer.deserialize(data["additionalProperties"])
        return types.MapType(
            key_type=key_type,
            value_type=value_type,
            nullable=data.get("nullable", False),
        )

    @classmethod
    @st.hookimpl
    def get_type_serializer(
        cls, type: types.Type, registry: "Registry"
    ) -> "TypeSerializer":
        if not isinstance(type, types.MapType):
            return None
        key_serializer = registry.get_type_serializer(type.key_type)
        value_serializer = registry.get_type_serializer(type.value_type)
        return cls(key_serializer, value_serializer)

    @classmethod
    @st.hookimpl
    def get_type_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> "TypeSerializer":
        if data.get("type") != "object" or "additionalProperties" not in data:
            return None
        key_type_data = data.get("keys", {"type": "string"})
        key_serializer = registry.get_type_serializer_from_data(key_type_data)
        value_type_data = data["additionalProperties"]
        value_serializer = registry.get_type_serializer_from_data(value_type_data)
        return cls(key_serializer, value_serializer)


@dc.dataclass(frozen=True)
class StructTypeSerializer(TypeSerializer):
    """
    Type serializer for structs
    """

    field_serializers: Dict[str, TypeSerializer]

    def serialize(self, type: types.Type) -> Any:
        fields = {}

        field_names = [field.name for field in type.fields]
        ordered_fields = sorted(self.field_serializers, key=field_names.index)

        required = []

        for key in ordered_fields:
            serializer = self.field_serializers[key]
            serialized = serializer.serialize(type[key].type)
            if not serialized.pop("nullable", False):
                required.append(key)
            fields[key] = serialized

        result = {
            "type": "object",
            "properties": fields,
            "required": required,
        }
        if type.nullable:
            result["nullable"] = True
        return result

    def deserialize(self, data: Any) -> types.Type:
        fields = []
        for key, field_type in data.get("properties", {}).items():
            field_type_dict = dict(
                field_type, nullable=key not in data.get("required", [])
            )
            serializer = self.field_serializers[key]
            fields.append(
                types.Field(name=key, type=serializer.deserialize(field_type_dict))
            )
        return types.StructType(
            fields=tuple(fields), nullable=data.get("nullable", False)
        )

    @classmethod
    @st.hookimpl
    def get_type_serializer(
        cls, type: types.Type, registry: "Registry"
    ) -> "TypeSerializer":
        if not isinstance(type, types.StructType):
            return None
        field_serializers = {}
        for field in type.fields:
            field_serializers[field.name] = registry.get_type_serializer(field.type)
        return cls(field_serializers)

    @classmethod
    @st.hookimpl
    def get_type_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> "TypeSerializer":
        if data.get("type") != "object" or "additionalProperties" in data:
            return None
        fields = data.get("properties", {})
        field_serializers = {}
        for key, field_type in fields.items():
            field_type_dict = dict(
                field_type, nullable=key not in data.get("required", [])
            )
            field_serializers[key] = registry.get_type_serializer_from_data(
                field_type_dict
            )
        return cls(field_serializers)


@dc.dataclass(frozen=True)
class NativeFunctionTypeSerializer(TypeSerializer):
    """
    Type serializer for function types
    """

    arg_serializers: Dict[str, TypeSerializer]
    return_type_serializer: TypeSerializer

    def serialize(self, type: types.Type) -> Any:
        fields = []

        field_names = [field.name for field in type.args]
        ordered_fields = sorted(self.arg_serializers, key=field_names.index)

        arg_types = {arg.name: arg.type for arg in type.args}

        for key in ordered_fields:
            serializer = self.arg_serializers[key]
            fields.append({"name": key, "type": serializer.serialize(arg_types[key])})

        result = {
            "type": "native_function",
            "args": fields,
            "return_type": self.return_type_serializer.serialize(type.return_type),
        }
        if type.nullable:
            result["nullable"] = True
        return result

    def deserialize(self, data: Any) -> types.Type:
        fields = []
        for field in data["args"]:
            serializer = self.arg_serializers[field["name"]]
            fields.append(
                types.Field(
                    name=field["name"], type=serializer.deserialize(field["type"])
                )
            )

        return_type = self.return_type_serializer.deserialize(data["return_type"])
        return types.NativeFunctionType(fields, return_type).with_nullable(
            data.get("nullable", False)
        )

    @classmethod
    @st.hookimpl
    def get_type_serializer(
        cls, type: types.Type, registry: "Registry"
    ) -> "TypeSerializer":
        if not isinstance(type, types.NativeFunctionType):
            return None
        field_serializers = {}
        for field in type.args:
            field_serializers[field.name] = registry.get_type_serializer(field.type)
        return_type_serializer = registry.get_type_serializer(type.return_type)
        return cls(field_serializers, return_type_serializer)

    @classmethod
    @st.hookimpl
    def get_type_serializer_from_data(
        cls, data: Any, registry: "Registry"
    ) -> "TypeSerializer":
        if data.get("type") != "native_function":
            return None
        field_serializers = {}
        for field in data["args"]:
            field_serializers[field["name"]] = registry.get_type_serializer_from_data(
                field["type"]
            )
        return_type_serializer = registry.get_type_serializer_from_data(
            data["return_type"]
        )
        return cls(field_serializers, return_type_serializer)


TYPE_SERIALIZER_CLASSES = [
    IntegerTypeSerializer,
    FloatTypeSerializer,
    BooleanTypeSerializer,
    StringTypeSerializer,
    AnyTypeSerializer,
    ArrayTypeSerializer,
    MapTypeSerializer,
    StructTypeSerializer,
    NativeFunctionTypeSerializer,
]


def register(registry: Optional["Registry"] = None) -> None:
    """
    Register default type serializer classes
    """
    if registry is None:
        registry = st.registry

    for cls in TYPE_SERIALIZER_CLASSES:
        registry.register(cls)



# import dataclasses as dc
# from typing import Type as PyType, Any, Optional, Dict


# @st.model
# class MyModel:
# 	"""

# 	"""
# 	a: int
# 	b: st.S.Struct["a": st.S.String]

# 	@st.method
# 	def method_one(self) -> int:
# 		"""
# 		blah blah
# 		"""
# 		return self

# 	@st.method(symbolic=True, property=True)
# 	def symbolic_method_one(obj):
# 		return obj.method_one() + obj.method_two()


# def model(cls: PyType[Any], dataclass_kwargs: Optional[Dict[str, Any]] = None) -> PyType[Any]:
# 	"""
# 	Wrap the given class to be a statey model whose hooks can be registered via
# 	registry.register(model_cls)
# 	"""
# 	if dataclass_kwargs is None:
# 		dataclass_kwargs = {}

# 	dataclass_cls = dc.dataclass(cls, frozen=True, **dataclass_kwargs)



"""
This module contains data structures that are only used in order to write
expressive code, and basically boil down to factories for the data structures
contained in the the other modules in this package
"""
import dataclasses as dc
import inspect
from functools import wraps
from typing import Any, Sequence, Dict, Callable, Type, Tuple, Union, Optional

import statey as st
from statey.syms import utils, types, func, impl, Object, path as path_mod, session
from statey.syms.plugins import ParseDataClassPlugin, EncodeDataClassPlugin


def function(
    func: Callable[[Any], Any],
    type: types.Type = utils.MISSING,
    registry: "Registry" = utils.MISSING,
) -> func.Function:
    """
    Construct a Function object for a regular python function
    """
    return utils.native_function(func, type, registry)


def map(
    func: Callable[[Any], Any], value: Any, return_type: types.Type = utils.MISSING
) -> Any:
    """
    Apply the function to the given object. If it is an object, convert `func` to a native
    function and apply it to the object's underlying value
    """
    if not isinstance(value, Object):
        return func(value)
    func_type = utils.single_arg_function_type(value._type, return_type)
    func_obj = function(func, func_type)
    return value._inst.map(func_obj)


def declarative(
    func: Callable[[Any], Any] = utils.MISSING,
    name_key: Callable[[str], str] = lambda x: x,
    ignore: Callable[[str], str] = lambda x: x.startswith("_"),
) -> Callable[[Any], Any]:
    """
    Wrap the given function as a function that declares statey objects.
    """

    def dec(_func):
        @wraps(_func)
        def wrapper(session, *args, name_key=name_key, ignore=ignore, **kwargs):
            @utils.scope_update_handler
            def update_handler(frame, name, value):
                absolute = name_key(name)
                # Ignore reassignments
                if (
                    ignore(name)
                    or isinstance(value, st.Object)
                    and isinstance(value._impl, impl.Reference)
                    and value._impl.path == absolute
                ):
                    return

                try:
                    typ = session.ns.resolve(absolute)
                except st.exc.SymbolKeyError:
                    annotations = frame.f_locals.get("__annotations__", {})
                    name_annotation = annotations.get(absolute, utils.MISSING)
                    frame.f_locals[name] = session[absolute:name_annotation] << value
                else:
                    session.set_data(absolute, value)
                    frame.f_locals[name] = session.ns.ref(absolute)

            update_handler(_func)(session, *args, **kwargs)
            return session

        return wrapper

    return dec if utils.is_missing(func) else dec(func)


@dc.dataclass(frozen=True)
class _FunctionFactory(utils.Cloneable):
    """
    An interface for creating function call objects
    """

    annotation: Any = utils.MISSING

    def __call__(self, func: Callable[[Any], Any]) -> "_FunctionFactoryWithFunction":
        typ = self.annotation
        if not utils.is_missing(typ):
            return_type = st.registry.get_type(typ)
            typ = utils.function_type(inspect.signature(func), return_type=return_type)
        return function(func, typ)

    def __getitem__(self, annotation: Any) -> "_FunctionFactory":
        # Use double square brackets to indicate the annotation is the function/callable
        if isinstance(annotation, list) and len(annotation) == 1:
            return self.clone(annotation=annotation[0])(annotation[0])
        return self.clone(annotation=annotation)


F = _FunctionFactory()


def autoencode(cls: Type[Any]) -> Type[Any]:
    """
    This method attempts to wrap the given class with a proper base class so that
    it will be deserialized properly when bieng added to the session
    """
    if isinstance(cls, type) and dc.is_dataclass(cls):
        st.registry.register(ParseDataClassPlugin(cls))
        st.registry.register(EncodeDataClassPlugin(cls))
        return cls
    raise NotImplementedError(f"No known encoding plugin implemented for {cls}.")


def filtered_type(data: Any, type: types.Type) -> types.Type:
    """
    Get a sub-structure of `type` given a set of input data
    """
    if isinstance(type, types.ArrayType):
        return [sub_structure(item, type.element_type) for item in data]
    if isinstance(type, types.StructType):
        fields = []
        for field in type.fields:
            if field.name in data:
                fields.append(field)
        return types.StructType(fields, type.nullable)
    return type


def join(head: Object, *tail: Sequence[Any]) -> Object:
    """
    Pass all of the given argument to a symbolic function that will just
    return the first element, but the result will depend on all of the
    additional arguments symbolically as well
    """
    head = Object(head)

    tail_type = st.registry.infer_type(tail)

    def join_func(left: head._type, right: tail_type) -> head._type:
        return left

    return utils.wrap_function_call(join_func, (head, tail)) >> head


def struct_replace(obj: Object, keep_type=True, **kwargs: Dict[str, Any]) -> Object:
    """
    Replace the given attributes in `object`
    """
    if isinstance(obj, types.Type):
        type_fields = []

        if not hasattr(obj, "fields"):
            raise TypeError(f"Invalid input type for replace(): {obj}")

        for field in obj.fields:
            if field.name in kwargs:
                continue
            type_fields.append(field)

        for key, value in kwargs.items():
            type_fields.append(types.Field(key, value))

        return obj.with_fields(type_fields)

    if not hasattr(obj._type, "fields"):
        raise TypeError(f"Invalid input type for replace(): {obj._type}")

    out = []
    for field in obj._type.fields:
        if field.name in kwargs:
            out.append((field.name, kwargs[field.name]))
        else:
            out.append((field.name, obj[field.name]))

    return Object(dict(out), obj._type) if keep_type else struct.new(out)


def struct_interpolate_one(
    obj: Object,
    path: str,
    new_value: Any,
    keep_type: bool = True,
    path_parser: Optional[path_mod.PathParser] = None,
) -> Object:
    """
    Interpolate a new value in a struct object or type, possibly
    nested
    """
    if path_parser is None:
        path_parser = path_mod.PathParser()

    comps = path_parser.split(path)
    if len(comps) > 1:
        head, *tail = comps
        rel_path = path_parser.join(tail)

        if isinstance(obj, types.Type):
            if not hasattr(obj, "fields"):
                raise TypeError(f"Invalid type for struct_interpolate: {obj}")

            field_type = obj[head].type
            new_type = struct_interpolate_one(
                obj=field_type,
                path=rel_path,
                new_value=new_value,
                path_parser=path_parser,
                keep_type=keep_type,
            )
            return struct_replace(obj, keep_type, **{head: new_type})

        field_value = obj[head]
        new_value = struct_interpolate_one(
            obj=field_value,
            path=rel_path,
            new_value=new_value,
            path_parser=path_parser,
            keep_type=keep_type,
        )
        return struct_replace(obj, keep_type, **{head: new_value})

    return struct_replace(obj, keep_type, **{path: new_value})


def struct_interpolate(
    obj: Object,
    new_values: Dict[str, Any],
    keep_type: bool = True,
    path_parser: Optional[path_mod.PathParser] = None,
) -> Object:
    """
    Extend struct_interpolate_one to allow providing a dictionary
    of new values
    """
    if path_parser is None:
        path_parser = path_mod.PathParser()

    out = obj
    for key, value in new_values.items():
        out = struct_interpolate_one(
            obj=out,
            path=key,
            new_value=value,
            keep_type=keep_type,
            path_parser=path_parser,
        )

    return out


def fill(
    obj: Object, output_type: types.StructType, get_value: Callable[[str], Any]
) -> Object:
    """
    Convert `obj` to `output_type` by filling missing fields with values resolved from get_value
    """
    if not hasattr(obj._type, "fields"):
        raise TypeError(f"Invalid input type for replace(): {obj._type}")

    if isinstance(get_value, Object):
        get_value = get_value.__getitem__

    fields = {field.name: field for field in obj._type.fields}

    out = {}
    for field in output_type.fields:
        if field.name in fields:
            out[field.name] = obj[field.name]
        else:
            out[field.name] = get_value(field.name)

    return Object(out, output_type, obj._registry)


def filter_struct(obj: Object, output_type: types.StructType):
    """
    fill() which will raise an error if any exist that appears in output_type
    doesn't appear in `obj`
    """

    def getter(name):
        raise ValueError(f"Unknown field {name}.")

    return fill(obj, output_type, getter)


def fill_unknowns(obj: Object, output_type: types.StructType) -> Object:
    """
    Convert `obj` to `output_type` by filling missing values with unknowns
    """
    unknown = Object(impl.Unknown(return_type=output_type), output_type, obj._registry)

    def get_value(name):
        return unknown[name]

    return fill(obj, output_type, get_value)


def ifnull(obj: Object, otherwise: Any) -> Object:
    """
    If the object is None, substitute the value from the other object
    """
    non_nullable = obj._type.with_nullable(otherwise is None)

    def _ifnull(first: obj._type) -> non_nullable:
        return otherwise if first is None else first

    _ifnull.__name__ = "ifnull"
    return function(_ifnull)(obj, otherwise)


def struct_drop(obj: Union[Object, types.Type], *fields: str) -> Object:
    """
    Drop the fields with the given names from `obj`
    """
    if isinstance(obj, types.Type):
        type_fields = []

        if not hasattr(obj, "fields"):
            raise TypeError(f"Invalid input type for replace(): {obj}")

        for field in obj.fields:
            if field.name in fields:
                continue
            type_fields.append(field)

        return st.StructType(type_fields, nullable=obj.nullable, meta=obj.meta)

    if not hasattr(obj._type, "fields"):
        raise TypeError(f"Invalid input type for replace(): {obj._type}")

    out = {}

    new_type = struct_drop(obj._type, *fields)
    for field in new_type.fields:
        out[field.name] = obj[field.name]

    res = Object(out, new_type)
    return res


def struct_add(
    obj: Union[Object, types.Type], *fields: Sequence[Tuple[str, Any, types.Type]]
) -> Object:
    """
    Add the given fields to `obj`
    """
    if isinstance(obj, types.Type):

        if not hasattr(obj, "fields"):
            raise TypeError(f"Invalid input type for replace(): {obj}")

        type_fields = list(obj.fields)

        for tup in fields:
            if len(tup) == 2:
                name, typ = tup
                typ = st.registry.get_type(typ)
                type_fields.append(types.Field(name, typ))
            else:
                raise ValueError(f"Invalid tuple length: {len(tup)} for {tup}.")

        return st.StructType(type_fields, nullable=obj.nullable, meta=obj.meta)

    if not hasattr(obj._type, "fields"):
        raise TypeError(f"Invalid input type for replace(): {obj._type}")

    type_fields = list(obj._type.fields)
    values = {field.name: obj[field.name] for field in fields}

    for tup in fields:
        if len(tup) == 2:
            name, obj = tup
            obj = st.registry.get_object(obj)
        elif len(tup) == 3:
            name, obj, typ = tup
            obj = Object(obj, typ)
        else:
            ValueError(f"Invalid tuple length: {len(tup)} for {tup}.")

        type_fields.append(types.Field(name, obj._type))
        values[name] = obj

    new_type = st.StructType(
        type_fields, nullable=obj._type.nullable, meta=obj._type.meta
    )
    return Object(values, new_type)


class _StructSymbolFactory:
    """
    Utility for simply create struct objects
    """

    def new(self, fields: Sequence[Tuple[str, Any]]) -> Object:
        out_fields = []
        for name, value in fields:
            out_fields.append(impl.StructField(name, Object(value)))
        return Object(impl.Struct(out_fields))

    def dict(self, data: Dict[str, Any]) -> Object:
        return self.new(data.items())

    def __call__(
        self, *args: Sequence[Tuple[str, Any]], **kwargs: Dict[str, Any]
    ) -> Object:
        """
        'call' factory method.
        """
        items = []
        items.extend(args)
        items.extend(kwargs.items())
        return self.new(args)

    def __getitem__(self, fields: Sequence[slice]) -> Object:
        """
        'getitem' interfce for creating structs using __getitem__ syntax
        """
        out_fields = []
        for slc in fields:
            out_fields.append((slc.start, slc.stop))
        return self.new(out_fields)


struct = _StructSymbolFactory()


def str(input: Any) -> Object:
    @function
    def str(input: st.Any) -> st.String:
        return __builtins__["str"](input)

    return str(input)


def int(input: Any) -> Object:
    @function
    def int(input: st.Any) -> st.Integer:
        return __builtins__["int"](input)

    return int(input)



import abc
import asyncio
import contextlib
import dataclasses as dc
import sys
from functools import reduce, partial
from itertools import chain
from typing import Optional, Sequence, Dict, Tuple, Any

import networkx as nx
from networkx.algorithms.dag import topological_sort, is_directed_acyclic_graph

from statey import exc
from statey.provider import Provider, ProviderId
from statey.resource import (
    ResourceSession,
    BoundState,
    ResourceState,
    Resource,
    ResourceGraph,
    StateConfig,
    StateSnapshot,
)
from statey.syms import session, types, Object, stack, utils
from statey.task import (
    TaskSession,
    SessionSwitch,
    GraphDeleteKey,
    GraphSetKey,
    create_task_session,
)


class PlanAction(abc.ABC):
    """
    A plan action is some subset of a full task graph for a given name.
    """

    @abc.abstractmethod
    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:
        """
        Add the tasks in this graph to the given graph. Note that the output
        node must be {prefix}:output
        """
        raise NotImplementedError

    @abc.abstractmethod
    def input_task(self, prefix: str) -> str:
        """
        Return the name of the task, given a prefix, that will be the "input" task when
        rendering this action
        """
        raise NotImplementedError

    @abc.abstractmethod
    def output_task(self, prefix: str) -> str:
        """
        Return the name of the task, given a prefix, that will be the "output" task when
        rendering this action
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class ExecuteTaskSession(PlanAction):
    """"""

    task_session: TaskSession
    task_input_key: str
    output_ref: Object
    graph_ref: Object
    output_key: str
    config_state: ResourceState
    previous_state: ResourceState
    resource: Resource
    input_symbol: Object

    def input_task(self, prefix: str) -> str:
        return f"{prefix}:input"

    def output_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:

        input_switch_task = SessionSwitch(
            input_session=output_session,
            input_symbol=self.input_symbol,
            output_session=self.task_session,
            output_key=self.task_input_key,
        )

        input_key = self.input_task(prefix)
        graph.add_node(input_key, task=input_switch_task, source=prefix)

        # If output state is null, our graph operation is a deletion. Otherwise, it's a "set"
        if self.config_state == self.resource.null_state:
            output_key = self.output_task(prefix)
            output_graph_task = GraphDeleteKey(
                key=self.output_key,
                resource_graph=resource_graph,
                input_session=self.task_session,
                input_symbol=self.graph_ref,
            )
            graph.add_node(output_key, task=output_graph_task, source=prefix)
        else:
            output_key = f"{prefix}:state"
            output_graph_task = GraphSetKey(
                input_session=self.task_session,
                input_symbol=self.graph_ref,
                dependencies=dependencies,
                key=self.output_key,
                resource_graph=resource_graph,
                state=self.config_state,
                finalize=self.resource.finalize,
            )
            graph.add_node(output_key, task=output_graph_task, source=prefix)

            output_switch_task = SessionSwitch(
                input_session=self.task_session,
                input_symbol=self.output_ref,
                output_session=output_session,
                output_key=self.output_key,
            )
            output_switch_key = self.output_task(prefix)
            graph.add_node(output_switch_key, task=output_switch_task, source=prefix)

            # We should always update the state before the session
            graph.add_edge(output_key, output_switch_key)

        task_graph = self.task_session.task_graph(resource_graph, self.output_key)
        if not task_graph.nodes:
            graph.add_edge(input_key, output_key)

        task_key = lambda x: f"{prefix}:task:{x}"

        for sub_node in topological_sort(task_graph):
            node_key = task_key(sub_node)
            task = task_graph.nodes[sub_node]["task"]

            graph.add_node(node_key, task=task, source=prefix)

            # If not predecessors, make it dependent on the input
            # switch
            if not task_graph.pred[sub_node]:
                graph.add_edge(input_key, node_key)
            # Otherwise, add upstream edges
            else:
                for pred in task_graph.pred[sub_node]:
                    graph.add_edge(task_key(pred), node_key)

            # The opposite of above--if no dependencies, make the
            # output a dependency
            if not task_graph.succ[sub_node]:
                graph.add_edge(node_key, output_key)


@dc.dataclass(frozen=True)
class SetValue(PlanAction):
    """
    Set a given value in either the resource graph, the output session, or both during execution
    """

    output_key: str
    output_symbol: Object
    set_graph: bool = True
    set_session: bool = True

    def __post_init__(self) -> None:
        if not (self.set_graph or self.set_session):
            raise ValueError(
                "At least one of `set_session` or `set_graph` must be true."
            )

    def graph_update_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def session_switch_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def input_task(self, prefix: str) -> str:
        if self.set_graph and self.set_session or not self.set_session:
            return self.graph_update_task(prefix)
        return self.session_switch_task(prefix)

    def output_task(self, prefix: str) -> str:
        if self.set_session:
            return self.session_switch_task(prefix)
        return self.graph_update_task(prefix)

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:

        graph_task_name = self.graph_update_task(prefix)
        if self.set_graph:
            graph_task = GraphSetKey(
                input_session=output_session,
                key=self.output_key,
                input_symbol=self.output_symbol,
                resource_graph=resource_graph,
                dependencies=dependencies,
            )
            graph.add_node(graph_task_name, task=graph_task, source=prefix)

        session_task_name = self.session_switch_task(prefix)
        if self.set_session:
            session_task = SessionSwitch(
                input_session=output_session,
                input_symbol=self.output_symbol,
                output_session=output_session,
                output_key=self.output_key,
            )
            graph.add_node(session_task_name, task=session_task, source=prefix)
            if self.set_graph:
                graph.add_edge(graph_task_name, session_task_name)


@dc.dataclass(frozen=True)
class ResourceSetValue(PlanAction):
    """
    SetValue for a resource, a no-op action
    """

    output_key: str
    output_symbol: Object
    resource: Resource
    state: ResourceState

    def graph_update_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def session_switch_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def input_task(self, prefix: str) -> str:
        return self.graph_update_task(prefix)

    def output_task(self, prefix: str) -> str:
        return self.session_switch_task(prefix)

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:

        session_task_name = self.session_switch_task(prefix)
        session_task = SessionSwitch(
            input_session=output_session,
            input_symbol=self.output_symbol,
            output_session=output_session,
            output_key=self.output_key,
        )
        graph.add_node(session_task_name, task=session_task, source=prefix)

        graph_task_name = self.graph_update_task(prefix)
        graph_task = GraphSetKey(
            input_session=output_session,
            key=self.output_key,
            input_symbol=self.output_symbol,
            resource_graph=resource_graph,
            dependencies=dependencies,
            state=self.state,
            finalize=self.resource.finalize,
        )
        graph.add_node(graph_task_name, task=graph_task, source=prefix)

        graph.add_edge(graph_task_name, session_task_name)


@dc.dataclass(frozen=True)
class DeleteValue(PlanAction):
    """"""

    delete_key: str

    def input_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def output_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:

        graph_task_name = self.input_task(prefix)
        graph_task = GraphDeleteKey(key=self.delete_key, resource_graph=resource_graph)
        graph.add_node(graph_task_name, task=graph_task, source=prefix)


@dc.dataclass(frozen=True)
class PlanNode:
    """
    A plan node contains information about the state of a given name and before
    and after this migration, including any unknowns. It also contains information
    about the up and downstream dependencies of the configuration and previous state
    respectively
    """

    key: str
    # This will always be fully resolved
    current_data: Any
    current_type: types.Type
    current_state: Optional[ResourceState]
    current_depends_on: Sequence[str]
    current_action: Optional[PlanAction]
    # This may contain unknowns
    config_data: Any
    config_type: types.Type
    config_state: Optional[ResourceState]
    config_depends_on: Sequence[str]
    config_action: Optional[PlanAction]
    enforce_dependencies: bool = True

    def input_task(self) -> Optional[str]:
        """"""
        return self.current_input_task()

    def output_task(self) -> Optional[str]:
        """"""
        return self.config_output_task()

    def current_prefix(self) -> str:
        """"""
        if self.current_action is not None and self.config_action is not None:
            return f"{self.key}:current"
        return self.key

    def config_prefix(self) -> str:
        """"""
        if self.current_action is not None and self.config_action is not None:
            return f"{self.key}:config"
        return self.key

    def current_output_task(self) -> Optional[str]:
        """
        The output task of migrating the current state. This may or may not be the
        same as output_task(); the case where it is the same is when either this key
        only exists in one of the previous of configured namespaces or when it is a single
        resource whose state is being migrated.
        """
        action = self.current_action or self.config_action
        if action is None:
            return None
        return action.output_task(self.current_prefix())

    def config_output_task(self) -> Optional[str]:
        """"""
        action = self.config_action or self.current_action
        if action is None:
            return None
        return action.output_task(self.config_prefix())

    def config_input_task(self) -> str:
        """"""
        action = self.config_action or self.current_action
        if action is None:
            return None
        return action.input_task(self.config_prefix())

    def current_input_task(self) -> str:
        """"""
        action = self.current_action or self.config_action
        if action is None:
            return None
        return action.input_task(self.current_prefix())

    def get_edges(
        self, other_nodes: Dict[str, "PlanNode"]
    ) -> Sequence[Tuple[str, str]]:
        """
        Get the edges for this node, given the other nodes.
        """
        edges = {}

        if not self.enforce_dependencies:
            return []

        # input_task = self.input_task()
        current_input_task = self.config_input_task()
        if current_input_task is None:
            return []

        config_input_task = self.config_input_task()
        current_output_task = self.current_output_task()
        config_output_task = self.config_output_task()

        for upstream in self.config_depends_on:
            node = other_nodes[upstream]
            ref = node.output_task()
            if ref is None:
                continue
            edges[ref, config_input_task] = {"optional": False}

        for downstream in self.current_depends_on:
            node = other_nodes[downstream]
            current_ref = node.current_output_task()
            config_ref = node.config_output_task()
            if config_ref is None:
                continue
            # So we are checking if our input/config currently depends on the _current_
            # output of the other task. If not, we'll make the other input dependent
            # on the _current output task_
            # if not {(current_ref, current_input_task)} & edges:
            #     input_ref = node.current_input_task()
            #     edges.add((current_output_task, input_ref))

            # if not {(config_ref, config_input_task), (current_ref, )} & edges:
            #     input_ref = node.config_input_task()
            #     edges.add((config_output_task, input_ref))

            check_set = (
                {(current_ref, current_input_task)}
                if self.current_action
                else {(config_ref, config_input_task)}
            )

            if not check_set & set(edges):
                input_ref = node.current_input_task()
                edges[current_output_task, input_ref] = {"optional": True}

        return [(*key, val) for key, val in edges.items()]

    def get_task_graph(
        self, resource_graph: ResourceGraph, output_session: session.Session
    ) -> nx.DiGraph:
        """
        Render a task graph for this specific node of the plan. The names specified
        by current_output_task() and output_task() should be tasks in this graph.
        """
        tasks = nx.DiGraph()

        if self.current_action is not None and self.config_action is not None:
            self.current_action.render(
                graph=tasks,
                prefix=self.current_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                # The current task will always have a null result
                dependencies=(),
            )
            self.config_action.render(
                graph=tasks,
                prefix=self.config_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                dependencies=self.config_depends_on,
            )
            # Make these two graphs depend on one another
            tasks.add_edge(self.current_output_task(), self.config_input_task())
        elif self.current_action:
            self.current_action.render(
                graph=tasks,
                prefix=self.current_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                dependencies=(),
            )
        elif self.config_action:
            self.config_action.render(
                graph=tasks,
                prefix=self.config_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                dependencies=self.config_depends_on,
            )

        return tasks


@dc.dataclass(frozen=True)
class Plan:
    """
    A plan contains enough information to inspect and execute the sequence
    of tasks to reach the state of config_session from state_graph
    """

    nodes: Sequence[PlanNode]
    providers: Sequence[Provider]
    config_session: ResourceSession
    state_graph: ResourceGraph
    task_graph: "ResourceTaskGraph" = dc.field(init=False, default=None)

    def check_dag(self, graph: nx.DiGraph, strict: bool = False) -> None:
        """
        Check that this graph is a DAG, attempting to remove optional
        edges to remove conflicts if necessary
        """
        while not strict:
            try:
                cycle = nx.find_cycle(graph, orientation="reverse")
            except nx.NetworkXNoCycle:
                return

            removed = False

            for left, right, _ in cycle:
                edge = graph.edges[left, right]
                if edge.get("optional"):
                    graph.remove_edge(left, right)
                    removed = True
                    break

            if not removed:
                break

        # This will raise a properly formatted error
        utils.check_dag(dag)

    def new_task_graph(self, strict: bool = False) -> "ResourceTaskGraph":
        """
        Render a full task graph from this plan
        """
        from statey.executor import ResourceTaskGraph

        state_graph = self.state_graph.clone()
        output_session = self.config_session.clone()

        node_dict = {node.key: node for node in self.nodes}

        graphs = []
        raw_edges = {}

        for node in self.nodes:
            graphs.append(
                node.get_task_graph(
                    resource_graph=state_graph, output_session=output_session
                )
            )
            node_edges = node.get_edges(node_dict)
            for edge_from, edge_to, data in node_edges:
                raw_edges.setdefault((edge_from, edge_to), []).append(data)

        edges = [
            (*key, {"optional": all(item.get("optional") for item in vals)})
            for key, vals in raw_edges.items()
        ]

        # This will raise an error if there are overlapping keys, which there
        # shouldn't be.
        full_graph = reduce(nx.union, graphs) if graphs else nx.DiGraph()
        full_graph.add_edges_from(edges)

        self.check_dag(full_graph, strict)

        return ResourceTaskGraph(full_graph, output_session, state_graph)

    def build_task_graph(self, force: bool = False, **kwargs) -> None:
        """
        Create the task graph for this plan if it does not already exist
        """
        if self.task_graph is not None and not force:
            return

        self.__dict__["task_graph"] = self.new_task_graph(**kwargs)

    def is_empty(self, include_metatasks: bool = False) -> bool:
        """
        Indicate whether this plan is empty, meaning any tasks it contains
        are metatasks or there are no tasks at all
        """
        self.build_task_graph()
        for node in self.task_graph.task_graph.nodes:
            task = self.task_graph.get_task(node)
            if include_metatasks or not task.is_metatask():
                return False
        return True


class Migrator(abc.ABC):
    """
    Migrator is the interface for creating plans for migrating one set of resource states
    to another
    """

    @abc.abstractmethod
    def create_task_session(self) -> TaskSession:
        """
        Create a new instance of a TaskSession, used by resource classes to coordinate one
        or more actions in a grap symbollically
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def plan(
        self,
        config_session: ResourceSession,
        state_session: Optional[ResourceSession] = None,
    ) -> Plan:
        """
        Create a task graph to create the resource states in config_session, migrating
        from state_session if it is passed
        """
        raise NotImplementedError


class DefaultMigrator(Migrator):
    """
    simple default Migrator implementation
    """

    def create_task_session(self) -> TaskSession:
        return create_task_session()

    async def plan_node(
        self,
        node: str,
        config_session: ResourceSession,
        state_graph: ResourceGraph,
        config_dep_graph: nx.DiGraph,
        state_dep_graph: nx.DiGraph,
        output_session: ResourceSession,
        providers: Dict[ProviderId, Provider],
    ) -> Sequence[PlanNode]:
        """"""
        previous_exists = node in state_dep_graph.nodes
        previous_state = None
        if node in state_dep_graph.nodes:
            previous_state = state_dep_graph.nodes[node]["state"]

        config_exists = node in config_dep_graph.nodes
        config_bound_state = None
        config_state = None
        try:
            config_bound_state = output_session.get_state(node)
        except exc.SymbolKeyError:
            pass
        else:
            config_state = config_bound_state.state

        current_depends_on = list(state_dep_graph.pred[node]) if previous_exists else []
        config_depends_on = list(config_dep_graph.pred[node]) if config_exists else []

        plan_nodes = []

        # We can handle this node with one action--otherwise we need two, though
        # one or both may just be state update operations.
        if (
            config_state is not None
            and previous_state is not None
            and config_state.resource == previous_state.resource
        ):
            provider = providers.get(config_state.provider)
            if provider is None:
                provider = providers[
                    config_state.provider
                ] = output_session.ns.registry.get_provider(*config_state.provider)
                await provider.setup()

            resource = provider.get_resource(config_state.resource)
            previous_resolved = state_dep_graph.nodes[node]["value"]

            config_partial_resolved = output_session.resolve(
                config_bound_state.input, decode=False, allow_unknowns=True
            )
            previous_snapshot = StateSnapshot(previous_resolved, previous_state)

            task_session_base = self.create_task_session()
            input_ref = task_session_base.ns.new("input", config_state.state.input_type)
            task_session_base.set_data("input", config_partial_resolved)

            task_session = task_session_base.clone()

            bound_config = StateConfig(input_ref, config_state)

            # A bunch of junk to handle errors nicely
            error, tb = None, None
            current_action = None
            try:

                try:
                    plan_output = await resource.plan(
                        previous_snapshot, bound_config, task_session
                    )
                except exc.NullRequired:
                    null_bound_config = StateConfig({}, resource.null_state)

                    old_task_session = self.create_task_session()
                    old_task_session.ns.new("input", resource.null_state.input_type)
                    old_task_session.set_data("input", {})
                    # old_task_session = task_session_base.clone()
                    og_plan_output = await resource.plan(
                        previous_snapshot, null_bound_config, old_task_session
                    )

                    if isinstance(og_plan_output, tuple) and len(og_plan_output) == 2:
                        og_output_ref, og_graph_ref = og_plan_output
                    else:
                        og_output_ref, og_graph_ref = og_plan_output, og_plan_output

                    null_snapshot = StateSnapshot({}, resource.null_state)
                    task_session = task_session_base.clone()
                    plan_output = await resource.plan(
                        null_snapshot, bound_config, task_session
                    )

                    current_action = ExecuteTaskSession(
                        task_session=old_task_session,
                        task_input_key="input",
                        output_key=node,
                        output_ref=og_output_ref,
                        graph_ref=og_graph_ref,
                        config_state=resource.null_state,
                        previous_state=previous_state,
                        resource=resource,
                        input_symbol=null_snapshot.obj,
                    )

            except Exception as err:
                _, _, tb = sys.exc_info()
                error = exc.ErrorDuringPlanningNode(
                    node, previous_snapshot, bound_config, err, tb
                )
            if error:
                try:
                    raise error.with_traceback(tb)
                except Exception:
                    stack.rewrite_tb(*sys.exc_info())
            # End error junk

            # Allow different refs for graph vs. session by returning a tuple of refs
            if isinstance(plan_output, tuple) and len(plan_output) == 2:
                output_ref, graph_ref = plan_output
            else:
                output_ref, graph_ref = plan_output, plan_output

            output_partial_resolved = task_session.resolve(
                output_ref, allow_unknowns=True, decode=False
            )
            output_session.set_data(node, output_partial_resolved)

            test_graph = task_session.task_graph(ResourceGraph(), node)
            # This indicates there are no tasks in the session.
            if not test_graph.nodes:
                try:
                    resolved_graph_ref = task_session.resolve(graph_ref, decode=False)
                # The graph reference isn't fully resolved yet, we still need
                # to execute the task session
                except exc.UnknownError:
                    pass
                else:
                    state_obj = Object(
                        resolved_graph_ref,
                        config_state.state.output_type,
                        config_session.ns.registry,
                    )
                    # Since we are not setting this key in the graph, we can
                    # act like the configuration has no dependencies during
                    # planning. This essentially lets us avoid circular
                    # dependencies because these no-op resources will never
                    # depend on anything.
                    plan_node = PlanNode(
                        key=node,
                        current_data=previous_resolved,
                        current_type=previous_state.state.output_type,
                        current_state=previous_state,
                        current_depends_on=current_depends_on,
                        current_action=None,
                        config_data=resolved_graph_ref,
                        config_type=config_state.state.output_type,
                        config_state=config_state,
                        config_depends_on=config_depends_on,
                        config_action=ResourceSetValue(
                            node, state_obj, resource, config_state
                        ),
                        enforce_dependencies=False,
                    )
                    plan_nodes.append(plan_node)
                    return plan_nodes

            action = ExecuteTaskSession(
                task_session=task_session,
                task_input_key="input",
                output_key=node,
                output_ref=output_ref,
                graph_ref=graph_ref,
                config_state=config_state,
                previous_state=previous_state,
                resource=resource,
                input_symbol=config_bound_state.input,
            )

            plan_node = PlanNode(
                key=node,
                current_data=previous_resolved,
                current_type=previous_state.state.output_type,
                current_state=previous_state,
                current_depends_on=current_depends_on,
                current_action=current_action,
                config_data=output_partial_resolved,
                config_type=config_state.state.output_type,
                config_state=config_state,
                config_depends_on=config_depends_on,
                config_action=action,
            )
            plan_nodes.append(plan_node)
            return plan_nodes

        args = {
            "key": node,
            "current_data": None,
            "current_type": None,
            "current_state": None,
            "current_depends_on": current_depends_on,
            "current_action": None,
            "config_data": None,
            "config_type": None,
            "config_state": None,
            "config_depends_on": config_depends_on,
            "config_action": None,
        }

        if previous_state:
            provider = providers.get(previous_state.provider)
            if provider is None:
                provider = providers[
                    previous_state.provider
                ] = config_session.ns.registry.get_provider(*previous_state.provider)
                await provider.setup()

            resource = provider.get_resource(previous_state.resource)
            previous_resolved = state_dep_graph.nodes[node]["value"]

            previous_snapshot = StateSnapshot(previous_resolved, previous_state)

            null_state = resource.null_state

            task_session = self.create_task_session()
            input_ref = task_session.ns.new("input", null_state.input_type)

            config_bound = StateConfig(input_ref, null_state)
            task_session.set_data("input", {})

            # A bunch of junk to handle errors nicely
            error, tb = None, None
            try:
                plan_output = await resource.plan(
                    previous_snapshot, config_bound, task_session
                )
            except Exception as err:
                _, _, tb = sys.exc_info()
                error = exc.ErrorDuringPlanningNode(
                    node, previous_snapshot, config_bound, err, tb
                )
            if error:
                try:
                    raise error.with_traceback(tb)
                except Exception:
                    stack.rewrite_tb(*sys.exc_info())
            # End error junk

            # Allow different refs for graph vs. session by returning a tuple of refs
            if isinstance(plan_output, tuple) and len(plan_output) == 2:
                output_ref, graph_ref = plan_output
            else:
                output_ref, graph_ref = plan_output, plan_output

            error, tb = None, None
            try:
                # In theory this should _not_ allow unknowns, but keeping it less strict for now.
                # Update: removed unknowns, they should not be needed
                # Update2: These are needed for now
                output_resolved = task_session.resolve(
                    output_ref, decode=False, allow_unknowns=True
                )
            except Exception as err:
                _, _, tb = sys.exc_info()
                error = exc.ErrorDuringPlanningNode(
                    node, previous_snapshot, config_bound, err, tb
                )
            if error:
                try:
                    raise error.with_traceback(tb)
                except Exception:
                    stack.rewrite_tb(*sys.exc_info())

            action = ExecuteTaskSession(
                task_session=task_session,
                task_input_key="input",
                output_key=node,
                output_ref=output_ref,
                graph_ref=graph_ref,
                config_state=resource.null_state,
                previous_state=previous_state,
                resource=resource,
                input_symbol=Object({}, config_bound.type, task_session.ns.registry),
            )

            args.update(
                {
                    "current_data": previous_resolved,
                    "current_type": previous_state.output_type,
                    "current_state": previous_state,
                    "current_action": action,
                }
            )
        elif previous_exists and not config_exists:
            args.update(
                {
                    "current_data": state_dep_graph.nodes[node]["value"],
                    "current_type": state_dep_graph.nodes[node]["type"],
                    "current_state": None,
                    "current_action": DeleteValue(node),
                    # If the old value is not stateful, there are no dependencies to deleting it.
                    "current_depends_on": [],
                }
            )

        if config_state:
            provider = providers.get(config_state.provider)
            if provider is None:
                provider = providers[
                    config_state.provider
                ] = config_session.ns.registry.get_provider(*config_state.provider)
                await provider.setup()

            resource = provider.get_resource(config_state.resource)
            config_partial_resolved = output_session.resolve(
                config_bound_state.input, decode=False, allow_unknowns=True
            )

            previous_snapshot = StateSnapshot({}, resource.null_state)

            task_session = self.create_task_session()
            input_ref = task_session.ns.new("input", config_state.state.input_type)
            task_session.set_data("input", config_partial_resolved)

            config_bound = StateConfig(input_ref, config_state)

            # A bunch of junk to handle errors nicely
            error, tb = None, None
            try:
                plan_output = await resource.plan(
                    previous_snapshot, config_bound, task_session
                )
            except Exception as err:
                _, _, tb = sys.exc_info()
                error = exc.ErrorDuringPlanningNode(
                    node, previous_snapshot, config_bound, err, tb
                )
            if error:
                try:
                    raise error.with_traceback(tb)
                except Exception:
                    stack.rewrite_tb(*sys.exc_info())
            # End error junk
            # Allow different refs for graph vs. session by returning a tuple of refs
            if isinstance(plan_output, tuple) and len(plan_output) == 2:
                output_ref, graph_ref = plan_output
            else:
                output_ref, graph_ref = plan_output, plan_output

            output_partial_resolved = task_session.resolve(
                output_ref, allow_unknowns=True, decode=False
            )

            output_session.set_data(node, output_partial_resolved)

            action = ExecuteTaskSession(
                task_session=task_session,
                task_input_key="input",
                output_key=node,
                output_ref=output_ref,
                graph_ref=graph_ref,
                config_state=config_state,
                previous_state=resource.null_state,
                resource=resource,
                input_symbol=config_bound_state.input,
            )

            args.update(
                {
                    "config_data": output_partial_resolved,
                    "config_type": config_state.state.output_type,
                    "config_state": config_state,
                    "config_action": action,
                }
            )
        elif config_exists:
            ref = config_session.ns.ref(node)
            action = SetValue(node, ref)

            args.update(
                {
                    "config_data": config_session.resolve(
                        ref, decode=False, allow_unknowns=True
                    ),
                    "config_type": config_session.ns.resolve(node),
                    "config_state": None,
                    "config_action": action,
                }
            )

        plan_node = PlanNode(**args)
        plan_nodes.append(plan_node)

        return plan_nodes

    async def plan(
        self,
        config_session: ResourceSession,
        state_graph: Optional[ResourceGraph] = None,
    ) -> Plan:

        from statey.executor import AsyncIOTaskGraph, AsyncIOGraphExecutor

        if state_graph is None:
            state_graph = ResourceGraph()

        # Ugly error handling :(
        error, tb = None, None
        try:
            config_dep_graph = config_session.dependency_graph()
        except Exception as err:
            _, _, tb = sys.exc_info()
            error = exc.ErrorDuringPlanning(err, tb)
        if error:
            try:
                raise error.with_traceback(tb)
            except Exception:
                stack.rewrite_tb(*sys.exc_info())

        state_dep_graph = state_graph.graph

        # For planning we need a copy of the config session that we will partially resolve
        # (maybe w/ unknowns) as we go
        output_session = config_session.clone()

        config_nodes = set(config_dep_graph.nodes)
        state_nodes = set(state_dep_graph.nodes)

        state_only_nodes = state_nodes - config_nodes

        plan_nodes = []
        providers = {}

        task_graph = nx.DiGraph()

        async def coro(node):
            plan_nodes.extend(
                await self.plan_node(
                    node=node,
                    config_session=config_session,
                    state_graph=state_graph,
                    config_dep_graph=config_dep_graph,
                    state_dep_graph=state_dep_graph,
                    output_session=output_session,
                    providers=providers,
                )
            )

        for node in chain(config_dep_graph.nodes, state_only_nodes):
            task_graph.add_node(node, coroutine_factory=partial(coro, node))

        task_graph.add_edges_from(((a, b) for a, b, _ in config_dep_graph.edges))
        task_graph_obj = AsyncIOTaskGraph(task_graph)
        executor = AsyncIOGraphExecutor()

        try:
            exec_info = await executor.execute_async(task_graph_obj)
        finally:
            providers = tuple(providers.values())
            await asyncio.gather(*(provider.teardown() for provider in providers))
        # Ugly error handling :(
        error, tb = None, None
        try:
            exec_info.raise_for_failure()
        except Exception as err:
            _, _, tb = sys.exc_info()
            error = exc.ErrorDuringPlanning(err, tb)
        if error:
            try:
                raise error.with_traceback(tb)
            except Exception:
                stack.rewrite_tb(*sys.exc_info())

        plan = Plan(tuple(plan_nodes), providers, config_session, state_graph)
        # This will ensure that the task graph is a valid DAG, not perfect but it will
        # do for now
        plan.build_task_graph()
        return plan



import abc
import dataclasses as dc
import inspect
from typing import Sequence, Dict, Any, Callable, Optional

from statey.syms import types, utils, Object, stack


class Function(abc.ABC):
    """
    A function takes some arguments and produces a result
    """

    type: types.FunctionType
    name: str

    def __call__(self, *args: Sequence[Any], **kwargs: Dict[str, Any]) -> Object:
        """
        Create a FunctionCall object by applying this function to the given arguments.
        """
        return utils.wrap_function_call(self, args, kwargs, frame_depth=1)

    @abc.abstractmethod
    def apply(self, arguments: Dict[str, Any]) -> Any:
        """
        Only API method for a function--actually apply the underlying implementation
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class NativeFunction(Function):
    """
    Regular python function
    """

    type: types.FunctionType
    func: Callable[[Any], Any]
    name: str = "<function>"

    @property
    def signature(self) -> Optional[inspect.Signature]:
        sig = self.__dict__.get("_signature", utils.MISSING)
        if utils.is_missing(sig):
            try:
                sig = inspect.signature(self.func)
            except TypeError:
                sig = None
            self.__dict__["_signature"] = sig
        return sig

    def apply(self, arguments: Dict[str, Any]) -> Any:
        if not self.signature:
            return self.func(**arguments)

        args = []
        kwargs = {}

        for name, param in self.signature.parameters.items():
            if name not in arguments:
                continue

            value = arguments[name]
            if param.kind == inspect._ParameterKind.POSITIONAL_ONLY:
                args.append(value)
            elif param.kind == inspect._ParameterKind.VAR_POSITIONAL:
                args.extend(value)
            elif param.kind == inspect._ParameterKind.VAR_KEYWORD:
                kwargs.update(value)
            else:
                kwargs[name] = value

        return self.func(*args, **kwargs)



import asyncio
import contextlib
import dataclasses as dc
import importlib
import json
import logging
import os
import sys
import types
from importlib.machinery import SourceFileLoader
from typing import Tuple, Optional, Sequence

import statey as st
from statey.executor import AsyncIOGraphExecutor
from statey.cli.inspect import Inspector
from statey.cli.log import ExecutorLoggingPlugin, CLILoggingHandler
from statey.cli.requirements import parse_requirements, parse_requirement

import click
import shutil


LOGGER = logging.getLogger(__name__)


class Controller:
    """
    Controller class for the statey CLI
    """

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        terminal_size: Optional[Tuple[int, int]] = None,
        registry: Optional[st.Registry] = None,
        fulltrace: bool = False,
        show_metatasks: bool = False,
        show_task_dag: bool = True,
        show_diff: bool = False,
        task_heartbeat: int = 15.0,
        conf_name: str = "statey_conf.py",
        env_prefix: str = "STATEY_",
        debug: bool = False,
    ) -> None:

        if registry is None:
            registry = st.registry

        if logger is None:
            logger = LOGGER

        if terminal_size is None:
            terminal_size = shutil.get_terminal_size((80, 20))

        self.logger = logger
        self.terminal_size = terminal_size
        self.fulltrace = fulltrace
        self.conf_name = conf_name
        self.env_prefix = env_prefix
        self.show_metatasks = show_metatasks
        self.task_heartbeat = task_heartbeat
        self.registry = registry
        self.show_task_dag = show_task_dag
        self.show_diff = show_diff
        self.debug = debug

        self.inspector = Inspector()

        self.state_manager = None
        self.resource_graph = None
        self.session = None
        self.conf_module = None
        self.module_name = None
        self.session_name = None
        self.plan = None
        self.plan_resource_graph = None
        self.exec_info = None
        self.logging_set_up = False

    def setup_logging(self) -> None:
        """"""
        if self.logging_set_up:
            return

        self.logger.setLevel(logging.DEBUG if self.debug else logging.INFO)
        self.logger.addHandler(CLILoggingHandler(self.fulltrace or self.debug))

    def setup_state_manager(self) -> None:
        """"""
        if self.state_manager is not None:
            return
        self.setup_config()
        try:
            self.state_manager = self.registry.get_state_manager()
        except Exception as err:
            self.logger.exception(
                "Error obtaining state manager: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

        self.logger.debug(
            "Successfully instantiated state manager: %r", self.state_manager
        )

    def setup_resource_graph(self) -> None:
        """"""
        if self.resource_graph is not None:
            return
        self.setup_state_manager()
        try:
            self.resource_graph = self.state_manager.load(self.registry)
        except Exception as err:
            self.logger.exception(
                "Error loading state: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

        self.logger.debug(
            "Successfully loaded resource graph with %r", self.state_manager
        )

    def import_module_from_name(
        self, name: str, descriptor: str = "module", allow_fail: bool = False
    ) -> types.ModuleType:
        """
        Import a module by name, w/ proper error handling
        """
        # First attempt to load the module
        # First attempt to load a module with this name. Failing that, try to load
        # the name as a source file
        module = None
        try:
            module = importlib.import_module(name)
        except ImportError:
            pass
        except Exception as err:
            self.logger.exception(
                "Error importing %s '%s': %s: %s",
                descriptor,
                name,
                type(err).__name__,
                err,
            )
            raise click.Abort from err
        else:
            self.logger.debug("Imported %s successfully.", name)

        if module is None:
            loader = SourceFileLoader("__main__", name)
            module = types.ModuleType(loader.name)
            try:
                loader.exec_module(module)
            except (ImportError, FileNotFoundError) as err:
                if not allow_fail:
                    self.logger.exception(
                        "Error importing %s: '%s': %s: %s",
                        descriptor,
                        name,
                        type(err).__name__,
                        err,
                    )
                    raise click.Abort from err
                return None
            except Exception as err:
                self.logger.exception(
                    "Error importing %s: '%s': %s: %s",
                    descriptor,
                    name,
                    type(err).__name__,
                    err,
                )
                raise click.Abort from err
            else:
                self.logger.debug("Imported %s successfully", name)

        return module

    def setup_config(self) -> None:
        """"""
        if self.conf_module is not None:
            return None

        module = self.import_module_from_name(self.conf_name, allow_fail=True)
        self.module_name = getattr(
            module,
            "MODULE_NAME",
            os.getenv(self.env_prefix + "MODULE_NAME", "statey_module.py"),
        )
        self.session_name = getattr(
            module,
            "SESSION_NAME",
            os.getenv(self.env_prefix + "SESSION_NAME", "module()"),
        )
        self.conf_module = module

        self.logger.debug(
            "Loading configuration module '%s' successfully.", self.conf_name
        )

    def setup_session(self) -> None:
        """"""
        if self.session is not None:
            return

        self.setup_config()
        module = self.import_module_from_name(self.module_name)

        if self.session_name.endswith("()"):
            value_name = self.session_name[:-2]
            try:
                session_factory = getattr(module, value_name)
            except AttributeError as err:
                self.logger.exception(
                    "Error loading session '%s': no value named %s in module %s",
                    self.session_name,
                    value_name,
                    self.module_name,
                )
                raise click.Abort from err

            session = st.create_resource_session()

            try:
                self.session = session_factory(session)
            except Exception as err:
                self.logger.exception(
                    "Error loading session '%s' from factory %r: %s: %s",
                    self.session_name,
                    session_factory,
                    type(err).__name__,
                    err,
                )
                raise click.Abort from err
            else:
                self.logger.debug(
                    "Loaded %s from %s successfully.",
                    self.session_name,
                    self.module_name,
                )

        else:
            try:
                self.session = getattr(module, self.session_name)
            except AttributeError as err:
                self.logger.exception(
                    "Error loading session '%s': no value named %s in module %s",
                    self.session_name,
                    self.session_name,
                    self.module_name,
                )
                raise click.Abort from err
            else:
                self.logger.debug(
                    "Loaded %s from %s successfully.",
                    self.session_name,
                    self.module_name,
                )

    def refresh_resource_graph(self, progressbar=True, **kwargs) -> None:
        """"""
        self.setup_resource_graph()
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(
                st.helpers.refresh(
                    self.resource_graph, progressbar=progressbar, **kwargs
                )
            )
        except Exception as err:
            self.logger.exception(
                "Error refreshing resource graph: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err
        else:
            self.logger.debug("Resource graph refreshed successfully.")

    def print_planning_execution_error(self, exc_type, exc_value, exc_tb) -> None:
        """"""
        exec_info = exc_value.exec_info
        tasks_by_status = exec_info.tasks_by_status()

        traces = {}
        for key in tasks_by_status.get(st.TaskStatus.FAILED, []):
            info = exec_info.task_graph.get_info(key)
            if info.error is not None:
                traces[key] = info.error

        trace_lines = []
        for key, error in traces.items():
            if self.fulltrace or self.debug:
                trace_lines.append(f"{key}:")
                trace_lines.append(error.format_exception())
                trace_lines.append("")
            else:
                trace_lines.append(f"{key}: {error.format_error_message()}")
                trace_lines.append("")

        trace_lines.insert(0, "")
        trace_str = "\n".join(trace_lines)

        self.logger.exception(
            "Planning failed due to encountering error(s). Error(s) by resource:\n%s",
            "\n".join(trace_lines),
            exc_info=(exc_type, exc_value, exc_tb),
        )

    def setup_plan(self) -> None:
        """"""
        if self.plan is not None:
            return

        self.setup_resource_graph()
        self.setup_session()

        loop = asyncio.get_event_loop()
        try:
            self.plan = loop.run_until_complete(
                st.helpers.plan(
                    session=self.session,
                    resource_graph=self.resource_graph,
                    refresh=True,
                    refresh_progressbar=True,
                )
            )
            self.plan_resource_graph = self.plan.task_graph.resource_graph
        except st.exc.ErrorDuringPlanning as err:
            if isinstance(err.exception, st.exc.ExecutionError):
                self.print_planning_execution_error(
                    type(err.exception), err.exception, sys.exc_info()[2]
                )
            else:
                self.logger.exception(
                    "Error occurred during planning: %s: %s",
                    type(err.exception).__name__,
                    err.exception,
                )
            raise click.Abort from err
        except Exception as err:
            self.logger.exception(
                "Error occurred during planning: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

    @contextlib.contextmanager
    def wrapped_providers_context(self):
        """"""
        self.setup_plan()

        ctx = st.helpers.providers_context(self.plan.providers)
        try:
            ctx.__enter__()
        except Exception as err:
            self.logger.exception(
                "Error setting up providers: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

        try:
            yield
        finally:
            try:
                ctx.__exit__(*sys.exc_info())
            except Exception as err:
                self.logger.exception(
                    "Error tearing down providers: %s: %s", type(err).__name__, err
                )
                raise click.Abort from err

    def execute_plan(self) -> None:
        """"""
        self.setup_plan()

        executor = AsyncIOGraphExecutor()
        executor.pm.register(
            ExecutorLoggingPlugin(self.show_metatasks, self.task_heartbeat)
        )

        with self.wrapped_providers_context():
            try:
                self.exec_info = executor.execute(self.plan.task_graph)
            except Exception as err:
                self.logger.exception(
                    "Unhandled error occurred during execution: %s: %s",
                    type(err).__name__,
                    err,
                )
                raise click.Abort from err

        # Right now plans can only be execute once, so get rid of it after execution
        self.plan = None

    def print_execution_summary(self) -> None:
        """
        Print out execution summary
        """
        if self.exec_info is None:
            raise ValueError("Plan has not been executed.")

        try:
            exec_summary = self.inspector.execution_summary(
                self.exec_info, self.show_metatasks
            )
            exec_summary_string = exec_summary.to_string(full_trace=self.fulltrace)
        except Exception as err:
            self.logger.exception(
                "Error generating execution summary: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

        self.logger.info(exec_summary_string)

    def print_plan_summary(self) -> None:
        """"""
        self.setup_plan()

        if self.plan.is_empty():
            self.logger.info(
                click.style("This plan is empty :)", fg="green", bold=True)
            )
            self.logger.info("")
            return

        try:
            plan_summary = self.inspector.plan_summary(self.plan, self.show_metatasks)
        except Exception as err:
            self.logger.exception(
                "Error generating plan summary: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

        if self.show_diff:
            try:
                plan_summary_string = plan_summary.to_string(self.terminal_size.columns)
            except Exception as err:
                self.logger.exception(
                    "Error generating plan summary string: %s: %s",
                    type(err).__name__,
                    err,
                )
                raise click.Abort from err

            self.logger.info(plan_summary_string)
            self.logger.info("")

        if self.show_task_dag:
            try:
                task_dag_string = plan_summary.task_dag_string()
            except Exception as err:
                self.logger.exception(
                    "Error generating task DAG string: %s: %s", type(err).__name__, err
                )
                raise click.Abort from err
            self.logger.info(
                f"{click.style('Task DAG', fg='green', bold=True)}:\n\n%s\n",
                task_dag_string,
            )

        self.logger.info("Resource summary: %s", plan_summary.short_summary_string())
        self.logger.info("")

    def dump_state(self) -> None:
        """"""
        self.setup_state_manager()

        resource_graph = self.resource_graph
        if self.plan_resource_graph is not None:
            resource_graph = self.plan_resource_graph

        try:
            self.state_manager.dump(resource_graph, self.registry)
        except Exception as err:
            self.logger.exception(
                "Error storing resource graph: %s: %s", type(err).__name__, err
            )
            raise click.Abort from err

    def load_provider(self, name: str) -> st.Provider:
        """"""
        try:
            return self.registry.get_provider(name)
        except Exception as err:
            self.logger.exception(
                "Error getting provider '%s': %s: %s", name, type(err).__name__, err
            )
            raise click.Abort from err

    def load_resource(self, provider: st.Provider, name: str) -> st.Resource:
        """"""
        try:
            return provider.get_resource(name)
        except Exception as err:
            self.logger.exception(
                "Error loading resource '%s' from provider '%s': %s: %s",
                name,
                provider.id.name,
                type(err).__name__,
                err,
            )
            raise click.Abort from err

    def print_resource_docs(self, resource: st.Resource) -> None:
        """"""
        name_style = {"fg": "green", "bold": True}

        output_type_serializer = st.registry.get_type_serializer(
            resource.UP.output_type
        )
        output_json = output_type_serializer.serialize(resource.UP.output_type)

        input_type_serializer = st.registry.get_type_serializer(resource.UP.input_type)
        input_json = input_type_serializer.serialize(resource.UP.input_type)

        click.echo(click.style("Name:", **name_style) + f" {resource.name}")
        click.echo(
            click.style("Inputs:", **name_style)
            + "\n"
            + json.dumps(input_json, indent=2, sort_keys=True)
        )
        click.echo(
            click.style("Outputs:", **name_style)
            + "\n"
            + json.dumps(output_json, indent=2, sort_keys=True)
        )

    def install_plugins(
        self,
        requirements: Sequence[str] = (),
        requirements_files: Sequence["file"] = (),
    ) -> None:
        """"""
        self.setup_config()

        installers = []
        plugins_by_installer = {}

        for requirement in requirements:
            parsed = parse_requirement(requirement)
            try:
                installer = self.registry.get_plugin_installer(requirement)
            except st.exc.NoPluginInstallerFound as err:
                self.logger.error(
                    "Unable to find installer for requirement %s", requirement
                )
                raise click.Abort from err

            try:
                idx = installers.index(installer)
            except ValueError:
                idx = len(installers)
                installers.append(installer)

            plugins_by_installer.setdefault(idx, []).append(parsed)

        for reqs_file in requirements_files:
            for requirement in parse_requirements(reqs_file):
                try:
                    installer = self.registry.get_plugin_installer(requirement.original)
                except st.exc.NoPluginInstallerFound as err:
                    self.logger.error(
                        "Unable to find installer for requirement: %s",
                        requirement.original,
                    )
                    raise click.Abort from err

                try:
                    idx = installers.index(installer)
                except ValueError:
                    idx = len(installers)
                    installers.append(installer)

                plugins_by_installer.setdefault(idx, []).append(requirement)

        @contextlib.contextmanager
        def installer_wrapper(installer):
            installer.setup()
            try:
                yield
            finally:
                installer.teardown()

        num_plugins = sum(map(len, plugins_by_installer.values()))
        self.logger.info("Installing %d plugin(s).", num_plugins)

        with contextlib.ExitStack() as stack:
            for idx, installer in enumerate(installers):
                stack.enter_context(installer_wrapper(installer))
                plugins = plugins_by_installer[idx]
                for plugin in plugins:
                    self.logger.info("Installing %s", plugin.original)
                    try:
                        installer.install(plugin.name, plugin.version)
                    except Exception as err:
                        self.logger.exception(
                            "Error encountered installing '%s': %s: %s",
                            plugin.original,
                            type(err).__name__,
                            err,
                        )

        self.logger.info("Installation completed successfully.")
